"""The splash submodule implements the SPLASH model :cite:p:`davis:2017a` for
calculating robust indices of radiation, evapotranspiration and plant-available soil
moisture.

It is based on the (SPLASH v1.0 implementation)[https://doi.org/10.5281/zenodo.376293]
published with the original description but has been extensively refactored and
restructured. The main changes are:

* The implementation now provides a programmatic API to the main model components,
  allowing the functionality to be imported and used within Python workflows. The
  original code was implemented as command line scripts that required editing for
  particular use cases.

* The code has been updated to accept numpy arrays and to calculate estimates
  simultaneously across arrays. The original code predominantly expected scalar inputs
  and required users to script iteratation over sites.

* Solar fluxes and the majority of evaporative fluxes do not vary given the initial data
  and so are now calculated once when a model instance is created. The original
  implementation recalculated many of these values when iterating over daily
  calculations, which reduces the memory footprint but at a substantial runtime cost.
  Only actual evapotranspiration (AET), soil moisture and runoff need to be calculated
  on a daily basis, because they depend on the soil moisture from the previous day.

* The hard-coded constants in the original code can now be modified by users. In
  particular, the maximum soil moisture capacity (``kWm``) was fixed globally in SPLASH
  v1.0 at 150mm: this can now be set by the user and can vary between sites.
"""  # noqa: D205



"""The :mod:`~pyrealm.pmodel.acclimation` module provides the
:class:`~pyrealm.pmodel.acclimation.AcclimationModel` class, which is a required input
to the :class:`~pyrealm.pmodel.pmodel.SubdailyPModel` class for fitting the P
Model at subdaily time scales.
"""  # noqa: D205

import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import interp1d  # type: ignore

from pyrealm.core.utilities import exponential_moving_average


class AcclimationModel:
    """Define the acclimation model to be used for the subdaily P Model.

    This class provides methods that allow data to be converted between
    photsynthetically 'fast' and 'slow' timescales. Many plant responses to changing
    enviroments are 'fast' - on the scale of minutes to seconds - but other responses
    are slow - over the scale of days or weeks, capturing the timescales over which
    plants will acclimate to changing conditions.

    An instance of this class is created using an array of :class:`numpy.datetime64`
    values that provide the observation datetimes for a dataset to be used in a
    :class:`~pyrealm.pmodel.pmodel.SubdailyPModel`. These datetimes need to be:

    * sampled at a subdaily frequency,
    * be strictly increasing,
    * be evenly spaced, using a spacing that evenly divides a day.

    The model can handle incomplete days at the start and end of the time series and
    will internally pad the datetimes to complete days in order to ensure correct
    sampling. The values in ``datetimes`` are assumed to be the precise times of the
    observations. If the datetimes represent the start or end of a sampling time span,
    then they should first be converted to a reasonable choice of observation time, such
    as the midpoint of the timespan.

    An acclimation window must then be set, defining a period of the day representing
    the environmental conditions that plants will acclimate to. This will typically be
    the time of day with highest productivity - usually around noon - when the light use
    efficiency of the plant can best make use of high levels of sunlight. This window is
    set using one of the following methods:

    * The :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.set_window` method sets a
      window centred on a given time during the day with a fixed width.
    * The :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.set_nearest` method sets
      the acclimation window as the single observation closest to a given time of day.
    * The :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.set_include` method
      allows the user to set an arbitrary selection of observations during the day as
      the acclimation window.

    Once a ``set_`` method has been applied then the other methods of the class can be
    used to:

    * Extract conditions within the acclimation window from time series for subdaily
      data  as either mean daily values
      (:meth:`~pyrealm.pmodel.acclimation.AcclimationModel.get_daily_means`) or as
      individual observations
      ( :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.get_window_values`)

    * Apply acclimation lags
      (:meth:`~pyrealm.pmodel.acclimation.AcclimationModel.apply_acclimation`) to
      generate daily realised values of acclimating variables
      from the daily optimal values.

    * Fill daily realised values for an acclimating variable back on to the subdaily
      scale
      (:meth:`~pyrealm.pmodel.acclimation.AcclimationModel.fill_daily_to_subdaily`),
      giving the actual acclimated responses used to calculate GPP at the subdaily
      scale.

    .. important::

        Many of the arguments to the ``AcclimationModel`` instance are used to control
        the behaviour of the methods described above. They are set as attributes of the
        ``AcclimationModel`` rather than as method arguments to ensure that the same
        settings are shared for all variables to which the given Acclimation model is
        applied.

        The argument description below describes the role of each setting but more
        details are also provided in the linked method descriptions.

    Args:
        datetimes: A sequence of datetimes for observations at a subdaily scale.
        allow_partial_data: Allows the
            :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.get_daily_means`
            method to switch between using :func:`numpy.mean` and :func:`numpy.nanmean`,
            so that daily mean values can be calculated even if the data in the
            acclimation window is incomplete.
        alpha: A float value that controls the speed of acclimation in the
            :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.apply_acclimation`
            method. The value must be in the range [0, 1], with smaller values giving
            slower acclimation and larger values giving quicker acclimation.
        allow_holdover: Allows the
            :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.apply_acclimation`
            method to adapt to missing values in the acclimation time series.
        update_point: Used in the
            :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.fill_daily_to_subdaily`
            method to set when plant behaviour updates to the conditions within the
            acclimation window. One of ``max`` (default) or ``mean``.
        fill_method: Used in the
            :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.fill_daily_to_subdaily`
            method to set the interpolation method used to fill subdaily observations.
            One of ``previous`` (default) or ``linear``.
    """

    def __init__(
        self,
        datetimes: NDArray[np.datetime64],
        allow_partial_data: bool = False,
        alpha: float = 1 / 15,
        allow_holdover: bool = False,
        update_point: str = "max",
        fill_method: str = "previous",
    ) -> None:
        # __init__ arguments
        self.datetimes: NDArray[np.datetime64]
        """The datetimes used to create the instance."""
        self.allow_partial_data: bool = allow_partial_data
        """Sets whether
        :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.get_daily_means` will ignore
        missing data."""
        self.alpha: float = alpha
        """Sets the speed of acclimation in
        :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.apply_acclimation`."""
        self.allow_holdover: bool = allow_holdover
        """Sets whether
        :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.apply_acclimation` will
        attempt to handle missing data."""
        self.fill_method: str = fill_method
        """The interpolation method to be used in 
        :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.fill_daily_to_subdaily`"""
        self.update_point: str = update_point
        """The update point to be used
        :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.fill_daily_to_subdaily`."""

        # Validation of __init__ arguments

        if not (0 <= self.alpha <= 1):
            raise ValueError("The alpha value must be in [0,1]")

        if self.update_point not in ("mean", "max"):
            raise ValueError(
                f"The update_point option must be one of "
                f"'mean' or 'max', not: '{self.update_point}'"
            )

        if self.fill_method not in ("linear", "previous"):
            raise ValueError(
                f"The fill_method option must be one of "
                f"'linear' or 'previous', not: '{self.fill_method}'"
            )

        # Attributes populated during initialisation by _validate_and_set_datetimes()
        self.spacing: np.timedelta64
        """The time interval between observations"""
        self.observation_dates: NDArray[np.datetime64]
        """The dates covered by the observations"""
        self.n_days: int
        """The number of days covered by the observations"""
        self.observation_times: NDArray[np.timedelta64]
        """The times of observations through the day as timedelta64 values in seconds"""
        self.n_obs: int
        """The number of daily observations"""
        self.padding: tuple[int, int]
        """The number of missing observations on the first and last days.

        Provides the number of observations to add to the start and end of the provided
        datetime sequence to give complete days."""
        self.padded_datetimes: NDArray[np.datetime64]
        """The datetime sequence padded to complete days."""

        # Run the initialisation logic steps
        self._validate_and_set_datetimes(datetimes=datetimes)

        # Attributes populated by the set_* methods

        self.include: NDArray[np.bool_]
        """Logical index of which daily observations are included in daily samples."""
        self.sample_datetimes: NDArray[np.datetime64]
        """Datetimes included in daily samples.

        This array is two-dimensional: the first axis is of length n_days and the second
        axis is of length n_samples.
        """
        self.sample_datetimes_mean: NDArray[np.datetime64]
        """The mean datetime of for each daily sample."""
        self.sample_datetimes_max: NDArray[np.datetime64]
        """The maximum datetime of for each daily sample."""

    def _validate_and_set_datetimes(self, datetimes: NDArray[np.datetime64]) -> None:
        """Validates the datetimes and sets key timing attributes.

        Args:
            datetimes: The datetimes to be validated and used to setup the model.
        """

        # The inputs must be:
        # - a one dimensional numpy datetime64 array
        # - stored with second precision
        # - with strictly increasing time deltas that are both evenly spaced and evenly
        #   divisible into a day
        if (
            not isinstance(datetimes, np.ndarray)
            or (datetimes.ndim > 1)
            or not np.issubdtype(datetimes.dtype, np.datetime64)
        ):
            raise ValueError(
                "Datetimes are not a 1 dimensional array with dtype datetime64"
            )

        # Update precision
        datetimes = datetimes.astype("datetime64[s]")

        # Check spacings between datetimes are all equal
        all_spacings = set(np.diff(datetimes))
        if len(all_spacings) > 1:
            raise ValueError("Datetime sequence not evenly spaced")

        spacing = all_spacings.pop()

        if spacing < 0:
            raise ValueError("Datetime sequence must be increasing")

        # Validate that we are dealing indeed with subdaily data
        if spacing > np.timedelta64(1, "D"):
            raise ValueError(
                "The AcclimationModel requires datetimes sampled at daily frequency "
                "or faster."
            )

        # Work out observations by date and look for incomplete dates at the start and
        # end of the time sequence
        dates = datetimes.astype("datetime64[D]")
        unique_dates, date_counts = np.unique(dates, return_counts=True)

        # Do we have at least three complete day to ensure we have a complete set of
        # observation times on day 2. Three days is not sensible for the model but this
        # ensures the padding can de determined.
        if len(unique_dates) < 3:
            ValueError("Not enough data to validate observation times")

        # Get the maximum number of observations per day and check it is evenly
        # divisible by the number of seconds in a day.
        obs_per_date = date_counts.max()
        day_remainder = (24 * 60 * 60) % obs_per_date
        if day_remainder:
            raise ValueError("Datetime spacing is not evenly divisible into a day")

        # Now populate attributes

        self.datetimes = datetimes
        self.spacing = spacing

        # Set the datetime padding
        self.padding = (
            obs_per_date - date_counts[0],
            obs_per_date - date_counts[-1],
        )

        # Get a complete set of observation times from day 2 - we have guaranteed
        # that the second day is complete.
        complete_times = datetimes[dates == unique_dates[1]]
        observation_timedeltas = complete_times - complete_times[0].astype(
            "datetime64[D]"
        )

        # Add padded datetimes as a time series containing only complete days, using an
        # offset to the last observations to avoid an off by one error from np.arange.
        self.padded_datetimes = np.arange(
            unique_dates[0] + observation_timedeltas[0],
            unique_dates[-1] + (observation_timedeltas[-1] + np.timedelta64(1, "s")),
            np.diff(observation_timedeltas)[0],
        )

        self.observation_dates = unique_dates
        self.n_days = len(self.observation_dates)
        self.observation_times = observation_timedeltas
        self.n_obs = len(self.observation_times)

    def _set_sampling_times(self) -> None:
        """Sets the times at which representative values are sampled.

        This private method must be called by all ``set_`` methods. It is used to
        update the instance to populate the following attributes:

        * :attr:`~pyrealm.pmodel.acclimation.AcclimationModel.sample_datetimes`: An
          array of the datetimes of observations included in daily samples of shape
          (n_day, n_sample).
        * :attr:`~pyrealm.pmodel.acclimation.AcclimationModel.sample_datetimes_mean`:
          An array of the mean daily datetime of observations included in daily samples.
        * :attr:`~pyrealm.pmodel.acclimation.AcclimationModel.sample_datetimes_max`:
          An array of the maximum daily datetime of observations included in daily
          samples.
        """

        # Get a view of the complete padded datetimes and then reshape along the first
        # axis into daily subarrays.
        times_by_day = self.padded_datetimes.view()
        times_by_day.shape = tuple([self.n_days, self.n_obs])

        # subset to the included daily values
        self.sample_datetimes = times_by_day[:, self.include]

        # Cannot use mean directly, so calculate the mean of the values expressed
        # as seconds and then add the mean values back onto the epoch.
        now = np.datetime64("now", "s")
        epoch = now - np.timedelta64(now.astype(int), "s")
        datetimes_as_seconds = self.sample_datetimes.astype(int)
        mean_since_epoch = datetimes_as_seconds.mean(axis=1).round().astype(int)
        max_since_epoch = datetimes_as_seconds.max(axis=1).round().astype(int)

        self.sample_datetimes_mean = epoch + mean_since_epoch.astype("timedelta64[s]")
        self.sample_datetimes_max = epoch + max_since_epoch.astype("timedelta64[s]")

    def set_window(
        self, window_center: np.timedelta64, half_width: np.timedelta64
    ) -> None:
        """Set the acclimation conditions to a daily time window.

        This method sets the daily values to sample using a time window, given the time
        of the window centre and its width. Both of these values must be provided as
        :class:`~numpy.timedelta64` values.

        Args:
            window_center: A timedelta since midnight to use as the window center
            half_width: A timedelta to use as a window width on each side of the center
        """

        if not (
            isinstance(window_center, np.timedelta64)
            and isinstance(half_width, np.timedelta64)
        ):
            raise ValueError(
                "window_center and half_width must be np.timedelta64 values"
            )

        # Find the timedeltas of the window start and end, using second resolution
        win_start = (window_center - half_width).astype("timedelta64[s]")
        win_end = (window_center + half_width).astype("timedelta64[s]")

        # Does that include more than one day?
        # NOTE - this might actually be needed at some point!
        if (win_start < 0) or (win_end > 86400):
            raise ValueError("window_center and half_width cover more than one day")

        # Now find which observation fall inclusively within that time window
        self.include = np.logical_and(
            win_start <= self.observation_times,
            win_end >= self.observation_times,
        )
        self.set_method = f"Window ({window_center}, {half_width})"
        self._set_sampling_times()

    def set_include(self, include: NDArray[np.bool_]) -> None:
        """Set the acclimation conditions using a logical array.

        This method sets which daily values will be sampled directly, by providing a
        boolean array for the daily observation times. The ``include`` array must be of
        the same length as the number of daily observations.

        Args:
            include: A boolean array indicating which daily observations to include.
        """

        if not (isinstance(include, np.ndarray) and include.dtype == np.bool_):
            raise ValueError("The include argument must be a boolean array")

        if self.n_obs != len(include):
            raise ValueError("The include array length is of the wrong length")

        self.include = include
        self.method = "Include array"
        self._set_sampling_times()

    def set_nearest(self, time: np.timedelta64) -> None:
        """Set the acclimation conditions to a single value closest to a target time.

        This method finds the daily observation time closest to a value provided as a
        :class:`~numpy.timedelta64` value since midnight. If the provided time is
        exactly between two observation times, the earlier observation will be used.
        The resulting single observation will then be used as the daily sample.

        Args:
            time: A :class:`~numpy.timedelta64` value.
        """

        if not isinstance(time, np.timedelta64):
            raise ValueError("The time argument must be a timedelta64 value.")

        # Convert to seconds and check it is in range
        time = time.astype("timedelta64[s]")
        if not (time >= 0 and time < 24 * 60 * 60):
            raise ValueError("The time argument is not >= 0 and < 24 hours.")

        # Calculate the observation time closest to the provided value
        nearest = np.argmin(abs(self.observation_times - time))
        include = np.zeros(self.n_obs, dtype=np.bool_)
        include[nearest] = True

        self.include = include
        self.method = f"Set nearest: {time}"
        self._set_sampling_times()

    def _raise_if_sampling_times_unset(self) -> None:
        """Check sampling times are set.

        This private method provides shared checking functionality for class methods
        that require the user to have run one of the ``set_`` methods to set sampling
        times.
        """

        if not hasattr(self, "include"):
            raise AttributeError(
                "Use a set_ method to select which daily observations "
                "are used for acclimation"
            )

    def _pad_values(self, values: NDArray[np.float64]) -> NDArray[np.float64]:
        """Pad values array to full days.

        This method takes an array representing daily values and pads the first and
        last day with NaN values so that they correspond to full days, similar to the
        datetimes array of this class.

        Args:
            values: An array containing the sample values. The first dimension should be
              matching the number of measurements, i.e., the first dimension of
              datetimes in :class:`~pyrealm.pmodel.acclimation.AcclimationModel`.
        """

        if self.padding == (0, 0):
            return values

        # Construct a padding iterable for np.pad, to pad only the first time axis
        padding_dims = list((self.padding,))
        padding_dims.extend([(0, 0)] * (values.ndim - 1))

        return np.pad(values, padding_dims, constant_values=(np.nan, np.nan))

    def get_window_values(self, values: NDArray[np.float64]) -> NDArray[np.float64]:
        """Extract acclimation window values for a variable.

        This method takes an array of values which has the same shape along the first
        axis as the datetimes used to create the instance and extracts the values from
        the acclimation window set using one of the ``set_`` methods.

        Args:
            values: An array of values for each observation.

        Returns:
            An array of the values within the defined acclimation window
        """

        self._raise_if_sampling_times_unset()

        # Check that the first axis has the same shape as the number of
        # datetimes in the init
        if values.shape[0] != self.datetimes.shape[0]:
            raise ValueError(
                "The first dimension of values is not the same length "
                "as the datetime sequence"
            )

        # Get a view of the values wrapped by date and then reshape along the first
        # axis into daily subarrays, leaving any remaining dimensions untouched.
        # When the values are padded the returned value is automatically a padded copy
        # of the original data, but if there is no padding, the original data is
        # returned and using a view and reshape should avoid copying the data.
        padded_values = self._pad_values(values)
        values_by_day = padded_values.view()
        values_by_day.shape = tuple([self.n_days, self.n_obs, *list(values.shape[1:])])

        # subset to the included daily values
        return values_by_day[:, self.include, ...]

    def get_daily_means(
        self,
        values: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Get the daily means of a variable during the acclimation window.

        This method extracts values from a given variable during a defined acclimation
        window set using one of the ``set_`` methods, and then calculates the daily mean
        of those values. The values can have any number of dimensions, but the first
        dimension must represent the time axis and have the same length as the original
        set of observation times.

        One attribute of the ``AcclimationModel`` class is used to tune the behaviour
        of this method:

        *  The ``allow_partial_data`` attribute allows the method to switch between
           using :func:`numpy.mean` and :func:`numpy.nanmean`, so that daily mean values
           can be calculated even if the data in the acclimation window is incomplete.
           Note that this will still return `np.nan` if _no_ data is present in the
           acclimation window.

        Args:
            values: An array of values to reduce to daily averages.

        Returns:
            An array of mean daily values during the acclimation window
        """

        self._raise_if_sampling_times_unset()

        daily_values = self.get_window_values(values)

        if self.allow_partial_data:
            return np.nanmean(daily_values, axis=1)

        return daily_values.mean(axis=1)

    def apply_acclimation(
        self,
        values: NDArray[np.float64],
        initial_values: NDArray[np.float64] | None = None,
    ) -> NDArray[np.float64]:
        r"""Apply acclimation to optimal values.

        Three key photosynthetic parameters (:math:`\xi`, :math:`V_{cmax25}` and
        :math:`J_{max25}`) show slow responses to changing environmental conditions and
        do not instantaneously adopt optimal values. This function applies exponential
        weighted averaging to the input values in order to calculate a lagged response.

        Two attributes of the ``AcclimationModel`` class are used to tune the behaviour
        of this method:

        * The ``alpha`` attribute controls the speed of acclimation. This value must be
          between 0 and 1 and is used as the weight term in an exponential moving
          average of daily values. Values closer to 1 lead to faster acclimation, with 1
          giving instantaneous acclimation and 0 giving no acclimation.

        * The ``allow_holdover`` attribute is used to allow the exponential weighted
          average function to handle missing data. Incomplete forcing data is common and
          both :math:`V_{cmax}` and :math:`J_{max}` are not estimable in some conditions
          (namely when :math:`m \le c^{\ast}`, see
          :class:`~pyrealm.pmodel.optimal_chi.OptimalChiPrentice14`) and so missing
          values in P Model predictions can arise even when the forcing data is
          complete. Since the weighted average process iterates over daily observations,
          it cannot normally be calculated with missing data but, with
          ``allow_holdover=True``, the underlying
          :func:`~pyrealm.core.utilities.exponential_moving_average` attempts to fill
          gaps within the daily time series.

        Args:
            values: An array of daily optimal values
            initial_values: Alternative starting values for the acclimated values
        """

        try:
            return exponential_moving_average(
                values=values,
                initial_values=initial_values,
                alpha=self.alpha,
                allow_holdover=self.allow_holdover,
            )
        except ValueError:
            raise ValueError(
                "Missing data in input values, try setting allow_holdover=True"
            )

    def fill_daily_to_subdaily(
        self,
        values: NDArray[np.float64],
        previous_values: NDArray[np.float64] | None = None,
    ) -> NDArray[np.float64]:
        """Resample daily variables onto the subdaily time scale.

        This method takes an array representing daily values and interpolates those
        values back onto the subdaily timescale used to create the
        :class:`~pyrealm.pmodel.acclimation.AcclimationModel` instance. The first
        axis of the `values` must be the same length as the number of days in the time
        series used to create the instance.

        Two attributes of the ``AcclimationModel`` class are used to tune the behaviour
        of this method:

        * The ``update_point`` attribute sets the point on the subdaily scale at which
          the plant updates acclimating variables to the new daily realised value. The
          default is ``max`` - the plant starts to acclimate to a new value at the end
          of the acclimation window - but this can also be set to ``mean`` to start
          acclimation from the middle of the acclimation window. Note that this setting
          implies the plant can predict the daily values between the mean and max
          observation time.

        * The ``fill_method`` attribute controls the interpolation used when filling
          data from daily to subdaily scales. Two interpolation kinds are currently
          implemented:

            1. The default option of ``previous`` interpolates the daily value as a
               constant, until updating to the next daily value. This option will fill
               values until the end of the time series.
            2. The alternative ``linear`` interpolates linearly between the update
               points of the daily values. The interpolated values are held constant for
               the first day and then interpolated linearly: this is to avoid plants
               adapting optimally to future conditions. Values will also not be filled
               beyond the end of the last window as the subsequent daily acclimated
               value is unknown.

        Subdaily observations before the update point on the first day of the time
        series are filled with ``np.nan``. The ``previous_values`` argument can be used
        to provide estimates of previous values of the variable at the start of the time
        series. These values can be used to avoid initial `np.nan` values and allow time
        series to be processed in blocks. This option is only currently implemented for
        interpolation using the ``fill_method='previous'`` option.

        Args:
            values: An array with the first dimension matching the number of days in the
               :class:`~pyrealm.pmodel.acclimation.AcclimationModel` instance.
            previous_values: An array of previous values from which to fill the
                variable.
        """

        interp_x_datetimes, interp_y_values, fill_value = (
            self._get_subdaily_interpolation_xy(
                values=values, previous_values=previous_values
            )
        )

        # Note that interp1d cannot handle datetime64 inputs, so need to interpolate
        # using datetimes cast to integer types
        interp_fun = interp1d(
            interp_x_datetimes.astype("int"),
            interp_y_values,
            axis=0,
            kind=self.fill_method,
            bounds_error=False,
            fill_value=fill_value,
            assume_sorted=True,
        )

        return interp_fun(self.datetimes.astype("int"))

    def _get_subdaily_interpolation_xy(
        self,
        values: NDArray[np.float64],
        previous_values: NDArray[np.float64] | None = None,
    ) -> tuple[
        NDArray[np.datetime64],
        NDArray[np.float64],
        tuple[NDArray[np.float64] | None, NDArray[np.float64] | None],
    ]:
        """Generate the interpolation data used to fill subdaily values.

        This private method is used to adjust the interpolation data used by
        :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.fill_daily_to_subdaily` to
        the interpolation settings for the class.

        It returns a tuple containing:

        * An array of the update times to be used for interpolation (the x values in the
          interpolation).
        * An array of the values to be used at those time (the y values in the
          interpolation), and
        * A tuple of two arrays giving the fill values to be applied at the start and
          end of the time axis.

        Args:
            values: An array with the first dimension matching the number of days in the
               :class:`~pyrealm.pmodel.acclimation.AcclimationModel` instance.
            previous_values: An array of previous values from which to fill the
                variable.
        """

        self._raise_if_sampling_times_unset()

        if values.shape[0] != self.n_days:
            raise ValueError(
                f"Acclimation model covers {self.n_days} days, input values has "
                f"length {values.shape[0]} on its first axis"
            )

        if self.update_point == "max":
            update_time = self.sample_datetimes_max
        elif self.update_point == "mean":
            update_time = self.sample_datetimes_mean

        # Check initial values settings - only allow with previous interpolation and
        # check the previous value shape matches
        if previous_values is not None:
            if self.fill_method == "linear":
                raise NotImplementedError(
                    "Using previous_values with fill_method='linear' is not implemented"
                )

            # Use np.broadcast_shapes here to handle checking array shapes. This is
            # mostly to catch the fact that () and (1,) are equivalent.
            try:
                np.broadcast_shapes(previous_values.shape, values.shape)
            except ValueError:
                raise ValueError(
                    f"The shape of previous_values {previous_values.shape} is not "
                    f"congruent with a time slice across the values {values[0].shape}"
                )

        # Use fill_value to handle extrapolation before or after update point:
        # - previous will fill the last value forward to the end of the time series,
        #   although this will be bogus if the interpolation time series extends beyond
        #   the next update point
        # - linear has a 1 day offset: the plant cannot adapt towards the next optimal
        #   value until _after_ the update point.

        if self.fill_method == "previous":
            # The fill values here are used to extend the last daily value out to the
            # end of the subdaily observations but also to fill any provided previous
            # values for subdaily observations _before_ the first daily value. If
            # the default previous value of None is supplied, this inserts np.nan as
            # expected.
            fill_value = (previous_values, values[-1])
        elif self.fill_method == "linear":
            # Shift the values forward by a day, inserting a copy of the first day at
            # the start. This then avoids plants seeing the future and provides values
            # up until the last observation.
            values = np.insert(values, 0, values[0], axis=0)
            update_time = np.append(
                update_time, update_time[-1] + np.timedelta64(1, "D")
            )
            fill_value = (np.array(np.nan), np.array(np.nan))

        return update_time, values, fill_value



"""The :mod:`~pyrealm.core.utilities` submodule provides shared utility functions
used to:

* check input arrays to functions have congruent shapes
* summarize object attributes in a tabular format
* share efficient implementations of core components, such as Horner-form polynomial
  calculation.
"""  # noqa: D205, D415

from collections.abc import Sequence

import numpy as np
import tabulate
from numpy.typing import NDArray


def check_input_shapes(*args: float | int | np.generic | np.ndarray | None) -> tuple:
    """Check sets of input variables have congruent shapes.

    This helper function validates inputs to check that they are either scalars or
    arrays and then that any arrays of the same shape. It returns a tuple of the common
    shape of the arguments, which is (1,) if all the arguments are scalar.

    Parameters:
        *args: A set of numpy arrays or scalar values

    Returns:
        The common shape of any array inputs or 1 if all inputs are scalar.

    Raises:
        ValueError: if the inputs contain arrays of differing shapes.

    Examples:
        >>> check_input_shapes(np.array([1,2,3]), 5)
        (3,)
        >>> check_input_shapes(4, 5)
        (1,)
        >>> check_input_shapes(np.array([1,2,3]), np.array([1,2]))
        Traceback (most recent call last):
        ...
        ValueError: Inputs contain arrays of different shapes.
    """

    # Collect the shapes of the inputs
    shapes = set()

    # DESIGN NOTES - currently allow:
    #   - scalars,
    #   - 0 dim ndarrays (also scalars but packaged differently)
    #   - 1 dim ndarrays with only a single value

    for val in args:
        if isinstance(val, np.ndarray):
            # Note that 0-dim ndarrays (which are scalars) pass through as do
            # one dimensional arrays with a single value (also a scalar)
            if val.size > 1:
                shapes.add(val.shape)
        # elif isinstance(val, Series):
        #    # Note that 0-dim ndarrays (which are scalars) pass through
        #    if val.ndim > 0:
        #        shapes.add(val.shape)
        elif val is None or isinstance(val, (float | int | np.generic)):
            pass  # No need to track scalars and optional values pass None
        else:
            raise ValueError(f"Unexpected input to check_input_shapes: {type(val)}")

    # shapes can be an empty set (all scalars) or contain one common shape
    # otherwise raise an error
    if len(shapes) > 1:
        raise ValueError("Inputs contain arrays of different shapes.")

    if len(shapes) == 1:
        return shapes.pop()

    return (1,)


def summarize_attrs(
    obj: object,
    attrs: tuple[tuple[str, str], ...],
    dp: int = 2,
    repr_head: bool = True,
) -> None:
    """Print a summary table of object attributes.

    This helper function prints a simple table of the mean, min, max and nan
    count for named attributes from an object for use in class summary
    functions.

    Args:
        obj: An object with attributes to summarize
        attrs: A tuple of 2-tuples giving attribute names and units.
        dp: The number of decimal places used in rounding summary stats.
        repr_head: A boolean indicating whether to show the object
            representation before the summary table.
    """

    # Check inputs
    if not isinstance(attrs, tuple):
        raise RuntimeError("attrs input not a tuple")

    # Create a list to hold variables and summary stats
    ret = []

    if len(attrs):
        # Process the attributes
        for attr, unit in attrs:
            data = getattr(obj, attr)

            # Avoid masked arrays - run into problems with edge cases with all NaN
            if isinstance(data, np.ma.core.MaskedArray):
                # Mypy complains about filled being untyped?
                data = data.filled(np.nan)  # type: ignore

            # Add the variable and stats to the list to be displayed
            attr_row = [
                attr,
                np.round(np.nanmean(data), dp),
                np.round(np.nanmin(data), dp),
                np.round(np.nanmax(data), dp),
                np.count_nonzero(np.isnan(data)),
            ]
            attr_row.append(unit)
            ret.append(attr_row)

    hdrs = ["Attr", "Mean", "Min", "Max", "NaN", "Units"]

    if repr_head:
        print(obj)

    print(tabulate.tabulate(ret, headers=hdrs))


def evaluate_horner_polynomial(
    x: NDArray[np.float64], cf: Sequence | NDArray
) -> NDArray[np.float64]:
    r"""Evaluates a polynomial with coefficients `cf` at `x` using Horner's method.

    Horner's method is a fast way to evaluate polynomials, especially for large degrees,
    that can be evaluated efficiently using the following rearrangement to avoid taking
    large powers.

    .. math::
        :nowrap:

        \[
            \begin{align*}
                p(x) &= 5 + 4x + 3x^2 + 2x^3\\
                    &= 5 + x(4 + x(3 + 2x))
            \end{align*}
        \]

    Args:
        x: The values at which to evaluate the polynomial
        cf: The coefficients of the polynomial, ordered from the
            lowest (constant) to the highest degree.
    """
    y = np.zeros_like(x)
    for c in reversed(cf):
        y = x * y + c
    return y


def exponential_moving_average(
    values: NDArray[np.float64],
    initial_values: NDArray[np.float64] | None = None,
    alpha: float = 0.067,
    allow_holdover: bool = False,
) -> NDArray[np.float64]:
    r"""Apply an exponential moving average to a variable.

    This function implements an exponential moving average for an input array, using
    the parameter `alpha` (:math:`\alpha`) to control the speed of convergence of the
    smoothed values (:math:`S`) to the input values (:math:`X`):

    .. math::

        S_{t} = S_{t-1}(1 - \alpha) + X_{t} \alpha

    For :math:`t_{0}`, the first value in the input values is used (:math:`S_{0} =
    X_{0}`), unless alternative initial values are provided. The ``values`` array can
    have multiple dimensions but the smoothing is only applied along the first
    dimension, which is assumed to represent time.

    By default, the ``values`` array must not contain any missing values (`numpy.nan`).
    However, when ``allow_holdover`` is set then the function attempts to handle missing
    data by holding over the last non-missing smoothed value. This obviously cannot
    replace missing data at the start of a sequence, but after this if the current input
    value is missing, then the previous smoothed value will be recycled until it can
    next be updated from the input data.

    +-------------------+--------+-------------------------------------------------+
    |                   |        |   Current input data (:math:`X_{t}`)            |
    +-------------------+--------+-----------------+-------------------------------+
    |                   |        |     NA          |   not NA                      |
    +-------------------+--------+-----------------+-------------------------------+
    | Previous          |    NA  |     NA          |     X_{t}                     |
    | smoothed data     +--------+-----------------+-------------------------------+
    | (:math:`S_{t-1}`) | not NA | :math:`S_{t-1}` | :math:`S_{t-1}(1-a) + X_{t}a` |
    +-------------------+--------+-----------------+-------------------------------+

    Args:
        values: The values to apply the memory effect to.
        initial_values: Values to use at :math:`S_{0}` in place of :math:`X_{0}`.
        alpha: The relative weight applied to the most recent observation.
        allow_holdover: Allow missing values to be filled by holding over earlier
            values.

    Returns:
        An array of the same shape as ``values`` with the memory effect applied.
    """

    # Check for nan and nan handling
    nan_present = np.any(np.isnan(values))
    if nan_present and not allow_holdover:
        raise ValueError("Missing values in data passed to exponential_moving_average")

    # Initialise the output storage and set the first values to be a slice along the
    # first axis of the input values
    smoothed_values = np.empty_like(values, dtype=np.float64)
    if initial_values is None:
        smoothed_values[0] = values[0]
    else:
        smoothed_values[0] = initial_values * (1 - alpha) + values[0] * alpha

    # Handle the data if there are no missing data,
    if not nan_present:
        # Loop over the first axis, in each case taking slices through the first axis of
        # the inputs. This handles arrays of any dimension.
        for idx in range(1, len(smoothed_values)):
            smoothed_values[idx] = (
                smoothed_values[idx - 1] * (1 - alpha) + values[idx] * alpha
            )

        return smoothed_values

    # Otherwise, do the same thing but handling missing data at each step.
    for idx in range(1, len(smoothed_values)):
        # Need to check for nan conditions:
        # - the previous value might be nan from an initial nan or sequence of nans, in
        #   which case the current value is accepted without weighting - it could be nan
        #   itself to extend a chain of initial nan values.
        # - the current value might be nan, in which case the previous value gets
        #   held over as the current value.
        prev_nan = np.isnan(smoothed_values[idx - 1])
        curr_nan = np.isnan(values[idx])
        smoothed_values[idx] = np.where(
            prev_nan,
            values[idx],
            np.where(
                curr_nan,
                smoothed_values[idx - 1],
                smoothed_values[idx - 1] * (1 - alpha) + values[idx] * alpha,
            ),
        )

    return smoothed_values



"""The pyrealm package.

TODO: complete this documentation
"""

import importlib.metadata
import os
import warnings

__version__ = importlib.metadata.version("pyrealm")


# Setup warnings to simpler one line warning
def warning_on_one_line(
    message: Warning | str,
    category: type[Warning],
    filename: str,
    lineno: int,
    line: str | None = None,
) -> str:
    """Provide a simple one line warning for use in docstrings."""
    filename = os.path.join("pyrealm", os.path.basename(filename))
    return f"{filename}:{lineno}: {category.__name__}: {message}\n"


warnings.formatwarning = warning_on_one_line



"""The ``evap`` submodule provides functions and classes to calculate evaporative
fluxes.
"""  # noqa: D205

from dataclasses import InitVar, dataclass, field

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst
from pyrealm.core.hygro import (
    calc_enthalpy_vaporisation,
    calc_psychrometric_constant,
    calc_saturation_vapour_pressure_slope,
)
from pyrealm.core.utilities import check_input_shapes
from pyrealm.core.water import calc_density_h2o
from pyrealm.splash.solar import DailySolarFluxes


@dataclass
class DailyEvapFluxes:
    """Calculate daily evaporative fluxes.

    This class calculates daily evaporative fluxes given temperature and atmospheric
    pressure for daily observations and the calculated solar fluxes for those
    observations. The class attributes provide components of those fluxes that depend
    only on the initial data and are not dependent on the previous model state when
    iterating over time.

    Two remaining components, the intersection hour angle (hi, degrees) and the
    estimated daily AET (aet_d, mm), depend on the soil moisture from the previous day
    and so must be calculated on a daily basis during either the spin process of the
    SPLASH model to estimate initial equilibrium soil moisture, or the daily calculation
    of soil moisture along a time series. These quantities are estimated using the
    :meth:`~pyrealm.splash.evap.DailyEvapFluxes.estimate_aet` method, given an estimate
    of soil moisture from the preceeding day.

    Args:
        solar: A :class:`~pyrealm.splash.solar.DailySolarFluxes` instance from which to
            calculate evaporative fluxes. See the class definition for the flux
            variables and units provided.
        kWm: The maximum soil water capacity (mm).
        tc: The air temperature of the observations (°C).
        pa: The atmospheric pressure of the observations (Pa).
        core_const: An instance of CoreConst.
    """

    solar: DailySolarFluxes
    pa: InitVar[NDArray]
    tc: InitVar[NDArray]
    kWm: NDArray[np.float64] = field(default_factory=lambda: np.array([150.0]))
    core_const: CoreConst = field(default_factory=lambda: CoreConst())

    sat: NDArray[np.float64] = field(init=False)
    """Slope of saturation vapour pressure temperature curve, Pa/K"""
    lv: NDArray[np.float64] = field(init=False)
    """Enthalpy of vaporization, J/kg"""
    pw: NDArray[np.float64] = field(init=False)
    """Density of water, kg/m^3"""
    psy: NDArray[np.float64] = field(init=False)
    """Psychrometric constant, Pa/K"""
    econ: NDArray[np.float64] = field(init=False)
    """Water-to-energy conversion factor"""
    cond: NDArray[np.float64] = field(init=False)
    """Daily condensation, mm"""
    eet_d: NDArray[np.float64] = field(init=False)
    """Daily equilibrium evapotranspiration (EET), mm"""
    pet_d: NDArray[np.float64] = field(init=False)
    """Daily potential evapotranspiration (PET), mm"""
    rx: NDArray[np.float64] = field(init=False)
    """Variable substitute, (mm/hr)/(W/m^2)"""

    def __post_init__(self, pa: NDArray[np.float64], tc: NDArray[np.float64]) -> None:
        """Calculate invariant components of evapotranspiration.

        The post_init method calculates the components of the evaporative fluxes that
        depend only on the initial data and are not dependent on the previous model
        state when iterating over time.
        """

        # Slope of saturation vap press temp curve, Pa/K
        self.sat = calc_saturation_vapour_pressure_slope(tc)

        # Enthalpy of vaporization, J/kg
        self.lv = calc_enthalpy_vaporisation(tc)

        # Density of water, kg/m^3
        self.pw = calc_density_h2o(tc, pa, core_const=self.core_const)

        # Psychrometric constant, Pa/K
        self.psy = calc_psychrometric_constant(tc, pa, core_const=self.core_const)

        # Calculate water-to-energy conversion (econ), m^3/J
        self.econ = self.sat / (self.lv * self.pw * (self.sat + self.psy))

        # Calculate daily condensation (cn), mm
        self.cond = (1e3) * self.econ * np.abs(self.solar.nighttime_net_radiation)

        # Estimate daily equilibrium evapotranspiration (eet_d), mm
        self.eet_d = (1e3) * self.econ * self.solar.daytime_net_radiation

        # Estimate daily potential evapotranspiration (pet_d), mm
        self.pet_d = (1.0 + self.core_const.k_w) * self.eet_d

        # Calculate variable substitute (rx), (mm/hr)/(W/m^2)
        self.rx = (3.6e6) * (1.0 + self.core_const.k_w) * self.econ

    def estimate_aet(
        self, wn: NDArray[np.float64], day_idx: int | None = None, only_aet: bool = True
    ) -> NDArray[np.float64] | tuple[NDArray, NDArray, NDArray]:
        """Estimate actual evapotranspiration.

        This method estimates the daily actual evapotranspiration (AET, mm/day), given
        estimates of the soil moisture  (``wn``) for observations. Optionally, the
        method can also return the the intersection hour angle (``hi``, degrees) and
        evaporative supply rate (``sw``, mm/h).

        By default, ``wn`` is expected to provide estimates for all observations across
        all days in the model, but ``day_idx`` can be set to indicate that ``wn`` is
        providing the soil moisture for one specific day across the observations.

        Args:
            wn: The soil moisture (mm).
            day_idx: An integer giving the index of the provided ``wn`` values along the
                time axis.
            only_aet: Should the function only return AET or AET, ``hi`` and ``sw``.

        Returns:
            An array of AET values or a tuple of arrays containing AET, ``hi`` and
            ``sw``.
        """

        # Check day_idx inputs and create the indexing object `didx`, used to either
        # subset the calculations to particular request days or use the entire array of
        # soil moisture. The slice here is used to programatically select `array[:]`.
        if day_idx is None:
            check_input_shapes(wn, self.sat)
            didx: int | slice = slice(self.sat.shape[0])
        else:
            check_input_shapes(wn, self.sat[day_idx])
            didx = day_idx

        # Calculate evaporative supply rate (sw), mm/h
        sw = self.core_const.k_Cw * wn / self.kWm

        # Validate evaporative supply rate
        if np.any(sw < 0):
            raise ValueError(
                "Evaporative supply rate contains values outside range of validity"
            )

        # Calculate the intersection hour angle (hi), degrees, guarding against np.nan
        # values from np.arccos(v > 1), setting this directly to 1
        hi_pre = (
            sw / (self.solar.rw[didx] * self.solar.rv[didx] * self.rx[didx])
            + self.solar.net_longwave_radiation[didx]
            / (self.solar.rw[didx] * self.solar.rv[didx])
            - self.solar.ru[didx] / self.solar.rv[didx]
        )
        hi = np.arccos(np.clip(hi_pre, -np.inf, 1)) / self.core_const.k_pir

        # Estimate daily actual evapotranspiration (aet_d), mm
        aet_d = (
            (sw * np.deg2rad(hi))
            + (
                self.rx[didx]
                * self.solar.rw[didx]
                * self.solar.rv[didx]
                * (
                    np.sin(np.deg2rad(self.solar.crossover_hour_angle[didx]))
                    - np.sin(np.deg2rad(hi))
                )
            )
            + (
                (
                    self.rx[didx] * self.solar.rw[didx] * self.solar.ru[didx]
                    - self.rx[didx] * self.solar.net_longwave_radiation[didx]
                )
                * (self.solar.crossover_hour_angle[didx] - hi)
                * self.core_const.k_pir
            )
        ) * (24.0 / np.pi)

        if only_aet:
            return aet_d
        else:
            return aet_d, hi, sw



r"""The module :mod:`~pyrealm.pmodel.optimal_chi` provides
the abstract base class :class:`~pyrealm.pmodel.optimal_chi.OptimalChiABC`,
which is used to support different implementations of the calculation of optimal chi.
"""  # noqa D210, D415

from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import PModelConst
from pyrealm.core.experimental import warn_experimental
from pyrealm.core.utilities import check_input_shapes, summarize_attrs
from pyrealm.pmodel.pmodel_environment import PModelEnvironment

OPTIMAL_CHI_CLASS_REGISTRY: dict[str, type[OptimalChiABC]] = {}
"""A registry for optimal chi calculation classes.

Different implementations of the calculation of optimal chi must all be subclasses of
:class:`~pyrealm.pmodel.optimal_chi.OptimalChiABC` abstract base class.
This dictionary is used as a registry for defined subclasses and a method name
is used to retrieve a particular implementation from this registry. For example:

.. code:: python

    prentice14_opt_chi = OPTIMAL_CHI_CLASS_REGISTRY['prentice14']
"""


class OptimalChiABC(ABC):
    r"""ABC for estimating optimal leaf internal :math:`\ce{CO2}` concentration.

    This provides an abstract base class for the implementation of alternative
    approaches to calculating the optimal :math:`\chi` and :math:`\ce{CO2}` limitation
    factors. All implementations estimate the following values, which are attributes of
    the resulting class instance.

    - The ratio of carboxylation to transpiration cost factors (``beta``,
      :math:`\beta`).
    - The variable ``xi`` (:math:`\xi`), which captures the sensitivity of :math:`\chi`
      to the vapour pressure deficit, and is related to the carbon cost of water (Medlyn
      et al. 2011; Prentice et 2014).
    - The optimal ratio of :math:`\ce{CO2}` partial pressures within the leaf (``ci``,
      :math:`c_i`) to the external environmental partial pressure (``ca```, :math:`c_i`)
      recorded. The optimal ratio ``chi`` (:math:`\chi = c_i/c_a`) is stored as well as
      the resulting internal partial pressure (``ci``).
    - The :math:`\ce{CO2}` limitation term for light-limited assimilation (``mj``,
      :math:`m_j`), the :math:`\ce{CO2}` limitation term for Rubisco-limited
      assimilation (``mc``, :math:`m_c`) and their ratio (``mjoc``, :math:`m_j/m_c`).

    The abstract base class requires that implementations of specific approaches define
    two methods:

    - `set_beta`: This method defines the calculation of the ``beta`` values to be used.
      The method is called by the ``__init__`` method when a subclass instance is
      created and should not change.
    - `estimate_chi`: This method defines the calculation of ``xi`` and then ``chi``,
      ``ci``, ``mj``, ``mc`` and ``mjoc``. This method is also called by the
      ``__init__`` method when a subclass instance is created, using the default defined
      calculation of ``xi``. However, it can also be called using modified values for
      ``xi`` to allow recalculation of these values. This is used primarily in fitting
      the P Model at subdaily scales, where the ``xi`` parameter acclimates slowly to
      changing environmental conditions.

    Args:
        env: An instance of
            :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment`  providing the
            photosynthetic environment for the model.
        pmodel_const: An instance of
            :class:`~pyrealm.constants.pmodel_const.PModelConst`.

    Returns:
        Instances of the abstract base class should not be created - use instances of
        specific subclasses.
    """

    method: str
    """A short method name used to identify the class in
    :data:`~pyrealm.pmodel.optimal_chi.OPTIMAL_CHI_CLASS_REGISTRY`.
    """
    is_c4: bool
    """A flag indicating if the method captures the C4 photosynthetic pathway."""
    requires: list[str]
    """A list of names of optional attributes of
    :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment` that must be populated
    to use a method.
    """

    def __init__(
        self,
        env: PModelEnvironment,
        pmodel_const: PModelConst = PModelConst(),
    ):
        self.env: PModelEnvironment = env
        """The PModelEnvironment containing the photosynthetic environment for the
        model."""
        self.shape: tuple[int, ...] = env.shape
        """The shape of the input environment data."""

        self.pmodel_const: PModelConst = pmodel_const
        """The PModelParams used for optimal chi estimation"""

        # Declare attributes populated by methods. These are typed but not assigned a
        # default value as they must be populated by the set_beta and estimate_chi
        # methods, which are called below, and so will be populated before __init__
        # returns.
        self.beta: NDArray[np.float64]
        """The ratio of carboxylation to transpiration cost factors."""
        self.xi: NDArray[np.float64]
        r"""Defines the sensitivity of :math:`\chi` to the vapour pressure deficit,
        related to the carbon cost of water (Medlyn et al. 2011; Prentice et 2014)."""
        self.chi: NDArray[np.float64]
        r"""The ratio of leaf internal to ambient :math:`\ce{CO2}` partial pressure
        (:math:`\chi`)."""
        self.mc: NDArray[np.float64]
        r""":math:`\ce{CO2}` limitation factor for RuBisCO-limited assimilation
        (:math:`m_c`)."""
        self.mj: NDArray[np.float64]
        r""":math:`\ce{CO2}` limitation factor for light-limited assimilation
        (:math:`m_j`)."""
        self.ci: NDArray[np.float64]
        r"""The leaf internal :math:`\ce{CO2}` partial pressure (:math:`c_i`)."""
        self.mjoc: NDArray[np.float64]
        r"""Ratio of :math:`m_j/m_c`."""

        # Run the calculation methods after checking for any required variables
        self._check_requires()
        self.set_beta()
        self.estimate_chi()

        # Validate that the subclass methods populate the attributes correctly.
        _ = check_input_shapes(
            env.ca, self.beta, self.xi, self.chi, self.ci, self.mj, self.mc, self.mjoc
        )

    @abstractmethod
    def set_beta(self) -> None:
        """Set the beta values."""

    @abstractmethod
    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate xi, chi and other variables."""

    def _check_requires(self) -> None:
        """Check additional required variables are present."""

        for required_var in self.requires:
            if not hasattr(self.env, required_var):
                raise ValueError(
                    f"{self.__class__.__name__} (method {self.method}) requires "
                    f"{required_var} to be provided in the PModelEnvironment."
                )

    def __repr__(self) -> str:
        """Generates a string representation of an OptimalChi instance."""
        return f"{type(self).__name__}(shape={self.shape})"

    def summarize(self, dp: int = 2) -> None:
        """Print OptimalChi summary.

        Prints a summary of the variables calculated within an instance
        of OptimalChi including the mean, range and number of nan values.

        Args:
            dp: The number of decimal places used in rounding summary stats.
        """

        attrs = (
            ("xi", " Pa ^ (1/2)"),
            ("chi", "-"),
            ("mc", "-"),
            ("mj", "-"),
            ("mjoc", "-"),
        )
        summarize_attrs(self, attrs, dp=dp)

    @classmethod
    def __init_subclass__(cls, method: str, is_c4: bool, requires: list[str]) -> None:
        """Initialise a subclass deriving from this ABC."""

        cls.method = method
        cls.is_c4 = is_c4
        cls.requires = requires
        OPTIMAL_CHI_CLASS_REGISTRY[cls.method] = cls


class OptimalChiPrentice14(
    OptimalChiABC,
    method="prentice14",
    is_c4=False,
    requires=[],
):
    r"""Calculate :math:`\chi` for C3 plants following :cite:`Prentice:2014bc`.

    Optimal :math:`\chi` is calculated following Equation 8 in
    :cite:`Prentice:2014bc`:

    .. math:: :nowrap:

        \[
            \begin{align*}
                \chi &= \Gamma^{*} / c_a + (1- \Gamma^{*} / c_a)
                    \xi / (\xi + \sqrt D ), \text{where}\\
                \xi &= \sqrt{(\beta (K+ \Gamma^{*}) / (1.6 \eta^{*}))}
            \end{align*}
        \]

    The :math:`\ce{CO2}` limitation term of light use efficiency (:math:`m_j`) is
    calculated following Equation 3 in :cite:`Wang:2017go`:

    .. math::

        m_j = \frac{c_a - \Gamma^{*}}
                {c_a + 2 \Gamma^{*}}

    Finally,  :math:`m_c` is calculated, following Equation 7 in :cite:`Stocker:2020dh`,
    as:

    .. math::

        m_c = \frac{c_i - \Gamma^{*}}{c_i + K},

    where :math:`K` is the Michaelis Menten coefficient of Rubisco-limited assimilation.

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]), 
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ... )
        >>> vals = OptimalChiPrentice14(env=env)
        >>> vals.chi.round(5)
        array([0.69435])
        >>> vals.mc.round(5)
        array([0.33408])
        >>> vals.mj.round(5)
        array([0.7123])
        >>> vals.mjoc.round(5)
        array([2.13211])
    """

    def set_beta(self) -> None:
        """Set ``beta`` to a constant C3 specific value."""
        # leaf-internal-to-ambient CO2 partial pressure (ci/ca) ratio
        self.beta = self.pmodel_const.beta_cost_ratio_c3

    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate ``chi`` for C3 plants."""

        if xi_values is not None:
            _ = check_input_shapes(self.env.ca, xi_values)
            self.xi = xi_values
        else:
            self.xi = np.sqrt(
                (self.beta * (self.env.kmm + self.env.gammastar))
                / (1.6 * self.env.ns_star)
            )

        self.chi = self.env.gammastar / self.env.ca + (
            1.0 - self.env.gammastar / self.env.ca
        ) * self.xi / (self.xi + np.sqrt(self.env.vpd))

        # Calculate m and mc and m/mc
        self.ci = self.chi * self.env.ca
        self.mj = (self.ci - self.env.gammastar) / (self.ci + 2 * self.env.gammastar)
        self.mc = (self.ci - self.env.gammastar) / (self.ci + self.env.kmm)
        self.mjoc = self.mj / self.mc


class OptimalChiPrentice14RootzoneStress(
    OptimalChiABC,
    method="prentice14_rootzonestress",
    is_c4=False,
    requires=["rootzonestress"],
):
    r"""Calculate :math:`\chi` for C3 plants following :cite:`Prentice:2014bc` with root
    zone stress penalty.

    The calculations are identical to
    :class:`~pyrealm.pmodel.optimal_chi.OptimalChiPrentice14` but apply an
    experimental rootzone stress penalty (:math:`f_{rz}`) to :math:`\beta` in the
    calculation of :math:`\xi`:

    .. math::

        \xi = \sqrt{(\beta f_{rz} (K+ \Gamma^{*}) / (1.6 \eta^{*}))}

    Calculation of the root zone stress factor is not currently part of the `pyrealm`
    package.

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ...     rootzonestress=0.5
        ... )
        >>> vals = OptimalChiPrentice14RootzoneStress(env=env)
        >>> vals.chi.round(5)
        array([0.62016])
    """  # noqa D210, D415 - long but sane title line.

    __experimental__ = True

    def set_beta(self) -> None:
        """Set ``beta`` to a constant C3 specific value."""

        # Warn that this is an experimental feature.
        warn_experimental("OptimalChiPrentice14RootzoneStress")

        # leaf-internal-to-ambient CO2 partial pressure (ci/ca) ratio
        self.beta = self.pmodel_const.beta_cost_ratio_c3

    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate ``chi`` for C3 plants."""

        if xi_values is not None:
            _ = check_input_shapes(self.env.ca, xi_values)
            self.xi = xi_values
        else:
            self.xi = np.sqrt(
                (
                    self.beta
                    * getattr(self.env, "rootzonestress")
                    * (self.env.kmm + self.env.gammastar)
                )
                / (1.6 * self.env.ns_star)
            )

        self.chi = self.env.gammastar / self.env.ca + (
            1.0 - self.env.gammastar / self.env.ca
        ) * self.xi / (self.xi + np.sqrt(self.env.vpd))

        # Calculate m and mc and m/mc
        self.ci = self.chi * self.env.ca
        self.mj = (self.ci - self.env.gammastar) / (self.ci + 2 * self.env.gammastar)
        self.mc = (self.ci - self.env.gammastar) / (self.ci + self.env.kmm)
        self.mjoc = self.mj / self.mc


class OptimalChiC4(
    OptimalChiABC,
    method="c4",
    is_c4=True,
    requires=[],
):
    r"""Estimate :math:`\chi` for C4 plants following :cite:`Prentice:2014bc`.

    Optimal :math:`\chi` is calculated as in
    :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiPrentice14`, but using a C4
    specific estimate of the unit cost ratio :math:`\beta`, see
    :attr:`~pyrealm.constants.pmodel_const.PModelConst.beta_cost_ratio_c4`.

    This method  sets :math:`m_j = m_c = m_{joc} = 1.0` to capture the boosted
    :math:`\ce{CO2}` concentrations at the chloropolast in C4 photosynthesis.

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ... )
        >>> vals = OptimalChiC4(env=env)
        >>> vals.chi.round(5)
        array([0.44967])
        >>> vals.mj.round(5)
        array([1.])
        >>> vals.mc.round(5)
        array([1.])
    """

    def set_beta(self) -> None:
        """Set ``beta`` to a constant C4 specific value."""
        # leaf-internal-to-ambient CO2 partial pressure (ci/ca) ratio
        self.beta = self.pmodel_const.beta_cost_ratio_c4

    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate ``chi`` for C4 plants, setting ``mj`` and ``mc`` to 1."""
        if xi_values is not None:
            _ = check_input_shapes(self.env.ca, xi_values)
            self.xi = xi_values
        else:
            self.xi = np.sqrt(
                (self.beta * (self.env.kmm + self.env.gammastar))
                / (1.6 * self.env.ns_star)
            )

        self.chi = self.env.gammastar / self.env.ca + (
            1.0 - self.env.gammastar / self.env.ca
        ) * self.xi / (self.xi + np.sqrt(self.env.vpd))

        self.ci = self.chi * self.env.ca

        # These values need to retain any dimensions of the original inputs - if
        # ftemp_kphio is set to 1.0 (i.e. no temperature correction) then the dimensions
        # of tc are lost.
        self.mc = np.ones(self.shape)
        self.mj = np.ones(self.shape)
        self.mjoc = np.ones(self.shape)


class OptimalChiC4RootzoneStress(
    OptimalChiABC,
    method="c4_rootzonestress",
    is_c4=True,
    requires=["rootzonestress"],
):
    r"""Estimate :math:`\chi` for C4 plants following :cite:`Prentice:2014bc` with
     root zone stress penalty.

    The calculations are identical to
    :class:`~pyrealm.pmodel.optimal_chi.OptimalChiC4` but apply an
    experimental rootzone stress penalty (:math:`f_{rz}`) to :math:`\beta` in the
    calculation of :math:`\xi`:

    .. math::

        \xi = \sqrt{(\beta f_{rz} (K+ \Gamma^{*}) / (1.6 \eta^{*}))}

    Calculation of the root zone stress factor is not currently part of the `pyrealm`
    package.

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ...     rootzonestress=0.5
        ... )
        >>> vals = OptimalChiC4RootzoneStress(env=env)
        >>> vals.chi.round(5)
        array([0.37659])
        >>> vals.mj.round(5)
        array([1.])
        >>> vals.mc.round(5)
        array([1.])
    """  # noqa D210, D415 - long but sane title line.

    def set_beta(self) -> None:
        """Set ``beta`` to a constant C4 specific value."""
        # leaf-internal-to-ambient CO2 partial pressure (ci/ca) ratio
        self.beta = self.pmodel_const.beta_cost_ratio_c4

    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate ``chi`` for C4 plants, setting ``mj`` and ``mc`` to 1."""
        if xi_values is not None:
            _ = check_input_shapes(self.env.ca, xi_values)
            self.xi = xi_values
        else:
            self.xi = np.sqrt(
                (
                    self.beta
                    * getattr(self.env, "rootzonestress")
                    * (self.env.kmm + self.env.gammastar)
                )
                / (1.6 * self.env.ns_star)
            )

        self.chi = self.env.gammastar / self.env.ca + (
            1.0 - self.env.gammastar / self.env.ca
        ) * self.xi / (self.xi + np.sqrt(self.env.vpd))

        self.ci = self.chi * self.env.ca

        # These values need to retain any dimensions of the original inputs - if
        # ftemp_kphio is set to 1.0 (i.e. no temperature correction) then the dimensions
        # of tc are lost.
        self.mc = np.ones(self.shape)
        self.mj = np.ones(self.shape)
        self.mjoc = np.ones(self.shape)


class OptimalChiLavergne20C3(
    OptimalChiABC,
    method="lavergne20_c3",
    is_c4=False,
    requires=["theta"],
):
    r"""Estimate :math:`\chi` for C3 plants using soil moisture corrected :math:`\beta`.

    Optimal :math:`\chi` is calculated using a definition of the unit cost ratio $\beta$
    as a function of soil moisture ($\theta$, m3 m-3), following :cite:`lavergne:2020a`:

    .. math:: :nowrap:

        \[
            \beta = e ^ {b\theta + a},
        \]

    The coefficients are experimentally derived values with defaults taken from
    Figure 6a of :cite:`lavergne:2020a` (:math:`a`,
    :attr:`~pyrealm.constants.pmodel_const.PModelConst.lavergne_2020_c3`;
    :math:`b`,
    :attr:`~pyrealm.constants.pmodel_const.PModelConst.lavergne_2020_c3`).

    Values of :math:`\chi` and other predictions are then calculated as in
    :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiPrentice14`. This method
    requires that `env` includes estimates of :math:`\theta`.

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ...     theta=np.array([0.5])
        ... )
        >>> vals = OptimalChiLavergne20C3(env=env)
        >>> vals.beta.round(5)
        array([224.75255])
        >>> vals.chi.round(5)
        array([0.73663])
        >>> vals.mc.round(5)
        array([0.34911])
        >>> vals.mj.round(5)
        array([0.7258])
        >>> vals.mjoc.round(5)
        array([2.07901])
    """

    def set_beta(self) -> None:
        """Set ``beta`` with soil moisture corrections."""

        # Calculate beta as a function of theta
        self.beta = np.exp(
            self.pmodel_const.lavergne_2020_c3[1] * getattr(self.env, "theta")
            + self.pmodel_const.lavergne_2020_c3[0]
        )

    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate ``chi`` for C3 plants."""

        if xi_values is not None:
            _ = check_input_shapes(self.env.ca, xi_values)
            self.xi = xi_values
        else:
            self.xi = np.sqrt(
                (self.beta * (self.env.kmm + self.env.gammastar))
                / (1.6 * self.env.ns_star)
            )

        self.chi = self.env.gammastar / self.env.ca + (
            1.0 - self.env.gammastar / self.env.ca
        ) * self.xi / (self.xi + np.sqrt(self.env.vpd))

        # Calculate m and mc and m/mc
        self.ci = self.chi * self.env.ca
        self.mj = (self.ci - self.env.gammastar) / (self.ci + 2 * self.env.gammastar)
        self.mc = (self.ci - self.env.gammastar) / (self.ci + self.env.kmm)
        self.mjoc = self.mj / self.mc


class OptimalChiLavergne20C4(
    OptimalChiABC,
    method="lavergne20_c4",
    is_c4=True,
    requires=["theta"],
):
    r"""Calculate soil moisture corrected :math:`\chi` for C4 plants.

    This method calculates :math:`\beta` as a function of soil moisture following
    the equation described in the
    :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiLavergne20C3` method.
    However,
    the default coefficients of the moisture scaling from :cite:`lavergne:2020a` for
    C3 plants are adjusted to match the theoretical expectation that :math:`\beta`
    for C4 plants is nine times smaller than :math:`\beta` for C3 plants (see
    :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiC4`): :math:`b`
    (:attr:`~pyrealm.constants.pmodel_const.PModelConst.lavergne_2020_c4`) is
    unchanged but
    :math:`a_{C4} = a_{C3} - log(9)`
    (:attr:`~pyrealm.constants.pmodel_const.PModelConst.lavergne_2020_c4`) .

    Following the calculation of :math:`\beta`, this method then follows the
    calculations described in
    :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiC4NoGamma`
    ::math:`m_j = 1.0`
    because photorespiration is negligible, but :math:`m_c` and hence
    :math:`m_{joc}` are calculated.

    .. NOTE::

        The research underlying :cite:`lavergne:2020a`, found **no relationship**
        between C4 :math:`\beta` values and soil moisture in leaf gas exchange
        measurements. This implementation asserts that there is a relationship and that
        the relationship is consistent with the patterns found for C3 plants.

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ...     theta=np.array([0.5])
        ... )
        >>> vals = OptimalChiLavergne20C4(env=env)
        >>> vals.beta.round(5)
        array([24.97251])
        >>> vals.chi.round(5)
        array([0.44432])
        >>> vals.mc.round(5)
        array([0.28091])
        >>> vals.mj.round(5)
        array([1.])
        >>> vals.mjoc.round(5)
        array([3.55989])
    """

    __experimental__ = True

    def set_beta(self) -> None:
        """Set ``beta`` with soil moisture corrections."""

        warn_experimental("OptimalChiLavergne20C4")

        # Calculate beta as a function of theta
        self.beta = np.exp(
            self.pmodel_const.lavergne_2020_c4[1] * getattr(self.env, "theta")
            + self.pmodel_const.lavergne_2020_c4[0]
        )

    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate ``chi`` for C4 plants excluding photorespiration."""

        # Calculate chi and xi as in Prentice 14 but removing gamma terms.
        if xi_values is not None:
            _ = check_input_shapes(self.env.ca, xi_values)
            self.xi = xi_values
        else:
            self.xi = np.sqrt((self.beta * self.env.kmm) / (1.6 * self.env.ns_star))

        self.chi = self.xi / (self.xi + np.sqrt(self.env.vpd))

        # mj is equal to 1 as gammastar is null
        self.mj = np.ones(self.shape)

        # Calculate m and mc and m/mc
        self.ci = self.chi * self.env.ca
        self.mc = (self.ci) / (self.ci + self.env.kmm)
        self.mjoc = self.mj / self.mc


class OptimalChiC4NoGamma(
    OptimalChiABC,
    method="c4_no_gamma",
    is_c4=True,
    requires=[],
):
    r"""Calculate optimal chi for C4 plants assuming negligible photorespiration.

    Calculates :math:`\chi` assuming that photorespiration (:math:`\Gamma^\ast`) is
    negligible for C4 plants. This simplifies the calculation of :math:`\xi` and
    :math:`\chi` compared to :meth:`~pyrealm.pmodel.optimal_chi.OptimalChiC4`,
    but uses the same C4 specific estimate of the unit cost ratio :math:`\beta`,
    :attr:`~pyrealm.constants.pmodel_const.PModelConst.beta_cost_ratio_c4`.

    .. math:: :nowrap:

        \[
            \begin{align*}
                \chi &= \xi / (\xi + \sqrt D ), \text{where}\\ 
                \xi &= \sqrt{(\beta K) / (1.6 \eta^{*}))}
            \end{align*}
        \]

    In addition, :math:`m_j = 1.0`  because photorespiration is negligible in C4
    photosynthesis, but :math:`m_c` and hence :math:`m_{joc}` are calculated, not
    set to one.

    .. math::

        m_c = \dfrac{\chi}{\chi + \frac{K}{c_a}}

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ... )
        >>> vals = OptimalChiC4NoGamma(env=env)
        >>> vals.chi.round(5)
        array([0.3919])
        >>> vals.mj.round(5)
        array([1.])
        >>> vals.mc.round(5)
        array([0.25626])
    """

    def set_beta(self) -> None:
        """Set constant ``beta`` for C4 plants."""

        # Calculate chi and xi as in Prentice 14 but removing gamma terms.
        self.beta = self.pmodel_const.beta_cost_ratio_c4

    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate ``chi`` for C4 plants excluding photorespiration."""

        # Calculate chi and xi as in Prentice 14 but removing gamma terms.
        if xi_values is not None:
            _ = check_input_shapes(self.env.ca, xi_values)
            self.xi = xi_values
        else:
            self.xi = np.sqrt((self.beta * self.env.kmm) / (1.6 * self.env.ns_star))

        # self.xi = np.sqrt(
        #     (self.beta * self.rootzonestress * self.env.kmm) /
        #     (1.6 * self.env.ns_star)
        # )

        self.chi = self.xi / (self.xi + np.sqrt(self.env.vpd))

        # mj is equal to 1 as gammastar is null
        self.mj = np.ones(self.shape)

        # Calculate m and mc and m/mc
        self.ci = self.chi * self.env.ca
        self.mc = (self.ci) / (self.ci + self.env.kmm)
        self.mjoc = self.mj / self.mc


class OptimalChiC4NoGammaRootzoneStress(
    OptimalChiABC,
    method="c4_no_gamma_rootzonestress",
    is_c4=True,
    requires=["rootzonestress"],
):
    r"""Calculate optimal chi for C4 plants assuming negligible photorespiration with
     root zone stress penalty.

    The calculations are identical to
    :class:`~pyrealm.pmodel.optimal_chi.OptimalChiC4NoGamma` but apply an
    experimental rootzone stress penalty (:math:`f_{rz}`) to :math:`\beta` in the
    calculation of :math:`\xi`:

    .. math::

        \xi = \sqrt{(\beta f_{rz} K) / (1.6 \eta^{*}))}

    Calculation of the root zone stress factor is not currently part of the `pyrealm`
    package.

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ...     rootzonestress=np.array([0.5])
        ... )
        >>> vals = OptimalChiC4NoGammaRootzoneStress(env=env)
        >>> vals.chi.round(5)
        array([0.31305])
        >>> vals.mj.round(5)
        array([1.])
        >>> vals.mc.round(5)
        array([0.21583])
    """  # noqa D210, D415 - long but sane title line.

    __experimental__ = True

    def set_beta(self) -> None:
        """Set constant ``beta`` for C4 plants."""

        warn_experimental("OptimalChiC4NoGammaRootzoneStress")

        # Calculate chi and xi as in Prentice 14 but removing gamma terms.
        self.beta = self.pmodel_const.beta_cost_ratio_c4

    def estimate_chi(self, xi_values: NDArray[np.float64] | None = None) -> None:
        """Estimate ``chi`` for C4 plants excluding photorespiration."""

        # Calculate chi and xi as in Prentice 14 but removing gamma terms.
        if xi_values is not None:
            _ = check_input_shapes(self.env.ca, xi_values)
            self.xi = xi_values
        else:
            self.xi = np.sqrt(
                (self.beta * getattr(self.env, "rootzonestress") * self.env.kmm)
                / (1.6 * self.env.ns_star)
            )

        self.chi = self.xi / (self.xi + np.sqrt(self.env.vpd))

        # mj is equal to 1 as gammastar is null
        self.mj = np.ones(self.shape)

        # Calculate m and mc and m/mc
        self.ci = self.chi * self.env.ca
        self.mc = (self.ci) / (self.ci + self.env.kmm)
        self.mjoc = self.mj / self.mc



"""The :mod:`~pyrealm.core.hygro` submodule provides conversion functions for
common hygrometric variables. The module provides conversions to vapour pressure
deficit, which is the required input for the
:class:`~pyrealm.pmodel.pmodel.PModel` from vapour pressure, specific humidity
and relative humidity. The implementation is drawn from and validated against the
``bigleaf`` R package.
"""  # noqa: D205

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst
from pyrealm.core.bounds import BoundsChecker
from pyrealm.core.utilities import evaluate_horner_polynomial


def calc_vp_sat(
    ta: NDArray[np.float64], core_const: CoreConst = CoreConst()
) -> NDArray[np.float64]:
    r"""Calculate vapour pressure of saturated air.

    This function calculates the vapour pressure of saturated air in kPa at a given
    temperature in °C, using the Magnus equation:

    .. math::

        P = a \exp\(\frac{b - T}{T + c}\),

    where :math:`a,b,c` are defined in
    :attr:`~pyrealm.constants.core_const.CoreConst.magnus_coef`.

    Args:
        ta: The air temperature in °C.
        core_const: An instance of :class:`~pyrealm.constants.core_const.CoreConst`
            giving the parameters for conversions.

    Returns:
        Saturated air vapour pressure in kPa.

    Examples:
        >>> # Saturated vapour pressure at 21°C
        >>> import numpy as np
        >>> temp = np.array([21])
        >>> calc_vp_sat(temp).round(6)
        array([2.480904])
        >>> from pyrealm.constants import CoreConst
        >>> allen = CoreConst(magnus_option='Allen1998')
        >>> calc_vp_sat(temp, core_const=allen).round(6)
        array([2.487005])
        >>> alduchov = CoreConst(magnus_option='Alduchov1996')
        >>> calc_vp_sat(temp, core_const=alduchov).round(6)
        array([2.481888])
    """

    # Magnus equation and conversion to kPa
    cf = core_const.magnus_coef
    vp_sat = cf[0] * np.exp((cf[1] * ta) / (cf[2] + ta)) / 1000

    return vp_sat


def convert_vp_to_vpd(
    vp: NDArray[np.float64],
    ta: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    """Convert vapour pressure to vapour pressure deficit.

    Args:
        vp: The vapour pressure in kPa
        ta: The air temperature in °C
        core_const: An instance of :class:`~pyrealm.constants.core_const.CoreConst`
            giving the settings to be used in conversions.

    Returns:
        The vapour pressure deficit in kPa

    Examples:
        >>> import numpy as np
        >>> from pyrealm.constants import CoreConst
        >>> vp = np.array([1.9])
        >>> temp = np.array([21])
        >>> convert_vp_to_vpd(vp, temp).round(7)
        array([0.5809042])
        >>> allen = CoreConst(magnus_option='Allen1998')
        >>> convert_vp_to_vpd(vp, temp, core_const=allen).round(7)
        array([0.5870054])
    """
    vp_sat = calc_vp_sat(ta, core_const=core_const)

    return vp_sat - vp


def convert_rh_to_vpd(
    rh: NDArray[np.float64],
    ta: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
    bounds_checker: BoundsChecker = BoundsChecker(),
) -> NDArray[np.float64]:
    """Convert relative humidity to vapour pressure deficit.

    Args:
        rh: The relative humidity (proportion in (0,1))
        ta: The air temperature in °C
        core_const: An instance of :class:`~pyrealm.constants.core_const.CoreConst`
            giving the settings to be used in conversions.
        bounds_checker: A BoundsChecker instance used to validate inputs.

    Returns:
        The vapour pressure deficit in kPa

    Examples:
        >>> import numpy as np
        >>> from pyrealm.constants import CoreConst
        >>> import sys; sys.stderr = sys.stdout
        >>> rh = np.array([0.7])
        >>> temp = np.array([21])
        >>> convert_rh_to_vpd(rh, temp).round(7)
        array([0.7442712])
        >>> allen = CoreConst(magnus_option='Allen1998')
        >>> convert_rh_to_vpd(rh, temp, core_const=allen).round(7)
        array([0.7461016])
        >>> rh_percent = np.array([70])
        >>> convert_rh_to_vpd(rh_percent, temp).round(7) #doctest: +ELLIPSIS
        pyrealm... contains values outside the expected range (0,1). Check units?
        array([-171.1823864])
    """

    rh = bounds_checker.check("rh", rh)

    vp_sat = calc_vp_sat(ta, core_const=core_const)

    return vp_sat - (rh * vp_sat)


def convert_sh_to_vp(
    sh: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    """Convert specific humidity to vapour pressure.

    Args:
        sh: The specific humidity in kg kg-1
        patm: The atmospheric pressure in kPa
        core_const: An instance of :class:`~pyrealm.constants.core_const.CoreConst`
            giving the settings to be used in conversions.

    Returns:
        The vapour pressure in kPa

    Examples:
        >>> import numpy as np
        >>> sh = np.array([0.006])
        >>> patm = np.array([99.024])
        >>> convert_sh_to_vp(sh, patm).round(7)
        array([0.9517451])
    """

    return sh * patm / ((1.0 - core_const.mwr) * sh + core_const.mwr)


def convert_sh_to_vpd(
    sh: NDArray[np.float64],
    ta: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    """Convert specific humidity to vapour pressure deficit.

    Args:
        sh: The specific humidity in kg kg-1
        ta: The air temperature in °C
        patm: The atmospheric pressure in kPa
        core_const: An instance of :class:`~pyrealm.constants.core_const.CoreConst`
            giving the settings to be used in conversions.

    Returns:
        The vapour pressure deficit in kPa

    Examples:
        >>> import numpy as np
        >>> from pyrealm.constants import CoreConst
        >>> sh = np.array([0.006])
        >>> temp = np.array([21])
        >>> patm = np.array([99.024])
        >>> convert_sh_to_vpd(sh, temp, patm).round(6)
        array([1.529159])
        >>> allen = CoreConst(magnus_option='Allen1998')
        >>> convert_sh_to_vpd(sh, temp, patm, core_const=allen).round(5)
        array([1.53526])
    """

    vp_sat = calc_vp_sat(ta, core_const=core_const)
    vp = convert_sh_to_vp(sh, patm, core_const=core_const)

    return vp_sat - vp


# The following functions are integrated from the evap.py implementation of SPLASH v1.


def calc_saturation_vapour_pressure_slope(
    tc: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Calculate the slope of the saturation vapour pressure curve.

    Calculates the slope of the saturation pressure temperature curve, following
    equation 13 of :cite:t:`allen:1998a`.

    Args:
        tc: The air temperature (°C)

    Returns:
        The calculated slope in kPa °C-1.
    """

    # TODO move these coefficients into constants?
    return (
        17.269
        * 237.3
        * 610.78
        * (np.exp(tc * 17.269 / (tc + 237.3)) / ((tc + 237.3) ** 2))
    )


def calc_enthalpy_vaporisation(tc: NDArray[np.float64]) -> NDArray[np.float64]:
    """Calculate the enthalpy of vaporization.

    Calculates the latent heat of vaporization of water as a function of
    temperature following :cite:t:`henderson-sellers:1984a`.

    Args:
        tc: Air temperature (°C)

    Returns:
        Calculated latent heat of vaporisation (J/Kg).
    """

    # TODO move these coefficients into constants?
    return 1.91846e6 * ((tc + 273.15) / (tc + 273.15 - 33.91)) ** 2


def calc_specific_heat(tc: NDArray[np.float64]) -> NDArray[np.float64]:
    """Calculate the specific heat of air.

    Calculates the specific heat of air at a constant pressure (:math:`c_{pm}`, J/kg/K)
    following :cite:t:`tsilingiris:2008a`. This equation is only valid for temperatures
    between 0 and 100 °C.

    Args:
        tc: Air temperature (°C)

    Returns:
        The specific heat of air values.
    """

    # TODO move these coefficients into constants?

    tc = np.clip(tc, 0, 100)
    cp = 1e3 * evaluate_horner_polynomial(
        tc,
        [
            1.0045714270,
            2.050632750e-3,
            -1.631537093e-4,
            6.212300300e-6,
            -8.830478888e-8,
            5.071307038e-10,
        ],
    )

    return cp


def calc_psychrometric_constant(
    tc: NDArray[np.float64], p: NDArray[np.float64], core_const: CoreConst = CoreConst()
) -> NDArray[np.float64]:
    r"""Calculate the psychrometric constant.

    Calculates the psychrometric constant (:math:`\lambda`, Pa/K) given the temperature
    and atmospheric pressure following :cite:t:`allen:1998a` and
    :cite:t:`tsilingiris:2008a`.

    Args:
        tc: Air temperature (°C)
        p: Atmospheric pressure (Pa)
        core_const: An instance of :class:`~pyrealm.constants.core_const.CoreConst`
            giving the settings to be used in conversions.

    Returns:
        The calculated psychrometric constant
    """

    # Calculate the specific heat capacity of water, J/kg/K
    cp = calc_specific_heat(tc)

    # Calculate latent heat of vaporization, J/kg
    lv = calc_enthalpy_vaporisation(tc)

    # Calculate psychrometric constant, Pa/K
    # Eq. 8, Allen et al. (1998)
    return cp * core_const.k_Ma * p / (core_const.k_Mv * lv)



"""This module defines a constant class for use with the two leaf, two stream model
:cite:t:`depury:1997a` of assimilation, implemented in the
:mod:`~pyrealm.pmodel.two_leaf` module.
"""  # noqa D210, D415

from dataclasses import dataclass, field

import numpy as np

from pyrealm.constants import ConstantsClass


@dataclass(frozen=True)
class TwoLeafConst(ConstantsClass):
    r"""Constants for the two leaf, two stream model.

    The derived constant for the reflectance of horizontal leaves (:math:`\rho_h`)
    is calculated automatically from the leaf scattering coefficient (:math:`\sigma`),
    following equation A20 of :cite:t:`depury:1997a`.

    .. math::

        \rho_h = \frac{1 - \sqrt{1 - \sigma}}{1 + \sqrt{1 - \sigma}}
    """

    atmospheric_scattering_coef: float = 0.426  # needs citation
    """Scattering coefficient of PAR in the atmosphere (:math:`f_a`, dimensionless)"""
    leaf_scattering_coef: float = 0.15
    r"""Scattering coefficient of PAR by leaves by reflection and transmission
    (:math:`\sigma`, dimensionless, Table 2. of :cite:t:`depury:1997a`)"""
    diffuse_reflectance: float = 0.036  # needs citation
    r"""Canopy reflection coefficient for diffuse PAR,(:math:`\rho_{cd}`,
    dimensionless)."""
    diffuse_extinction_coef: float = 0.719  # needs citation
    """Diffuse and scattered diffuse PAR extinction coefficient (:math:`k_d'`,
    dimensionless)."""
    solar_obscurity_angle: float = np.pi / 180  # 1 degree in rads, needs a citation
    r"""Solar obscurity angle (:math:`\beta_{ob}` radians)"""
    direct_beam_extinction_numerator: float = 0.5
    r"""Numerator of the extinction beam coefficient calculation (:math:`n`) for direct
     light."""
    scattered_beam_extinction_numerator: float = 0.46
    r"""Numerator of the extinction beam coefficient calculation (:math:`n`) for
     scattered light."""
    atmos_transmission_par: float = 0.72
    """The atmospheric transmission coefficient of photosynthetically active radiation,
    used in the calculation of fraction of diffuse light reaching the canopy
    (:math:`a`, Table 5. of :cite:t:`depury:1997a`)."""
    vcmax_lloyd_coef: tuple[float, float] = (0.00963, 2.43)
    """Coefficients of the canopy extinction coefficient (:math:`k_v`) function, taken
    from Figure 10 of :cite:`lloyd:2010a`"""
    jmax25_wullschleger_coef: tuple[float, float] = (29.1, 1.64)
    r"""Coefficients of the empirical relationship between the maximum rate of electron
     transport (:math:`J_{max25}`) and the carboxylation rate (:math:`V_{cmax25}`),
     values taken from :cite:`wullschleger:1993a`."""

    horizontal_leaf_reflectance: float = field(init=False)
    r"""The reflectance coefficient for horizontal leaves (:math:`\rho_h`)."""

    def __post_init__(self) -> None:
        """Set derived constants.

        This requires setattr because the dataclass is frozen.
        """

        object.__setattr__(
            self,
            "horizontal_leaf_reflectance",
            (1 - np.sqrt(1 - self.leaf_scattering_coef))
            / (1 + np.sqrt(1 - self.leaf_scattering_coef)),
        )



"""The :mod:`~pyrealm.pmodel` module includes the following submodules:

* :mod:`~pyrealm.pmodel.pmodel`,
* :mod:`~pyrealm.pmodel.pmodel_environment`,
* :mod:`~pyrealm.pmodel.optimal_chi` and
* :mod:`~pyrealm.pmodel.jmax_limitation` provide classes implementing the core
  calculations of the P Model.
* The :mod:`~pyrealm.pmodel.functions` submodule provides key standalone functions for
  those calculations.
* The :mod:`~pyrealm.pmodel.isotopes` submodule provides a class to estimate isotopic
  discrimination within the P Model.
* The :mod:`~pyrealm.pmodel.competition` submodule provides a competition model for the
  C3 and C4 photosynthetic pathways.

Note that the documentation of functions and methods includes two lists of parameters:

Parameters
  These are the arguments specific to the class, method or function signature.

Constants
  These are shared parameters of the PModel , which are taken from the
  :class:`~pyrealm.constants.pmodel_const.PModelConst` dataclass. These can be changed
  by the user but are typically used to configure an entire analysis rather than a
  single function.
"""  # noqa: D415

# This __init__ file imports the following members from submodules in order to
# flatten the namespace for the main public components and setup.cfg applies
# # to the whole file.

from pyrealm.pmodel.acclimation import AcclimationModel
from pyrealm.pmodel.competition import (
    C3C4Competition,
    calculate_tree_proportion,
    convert_gpp_advantage_to_c4_fraction,
)
from pyrealm.pmodel.functions import (
    calc_co2_to_ca,
    calc_ftemp_inst_rd,
    calc_gammastar,
    calc_kmm,
    calc_ns_star,
    calc_soilmstress_mengoli,
    calc_soilmstress_stocker,
    calc_viscosity_h2o,
    calculate_kattge_knorr_arrhenius_factor,
    calculate_simple_arrhenius_factor,
)
from pyrealm.pmodel.isotopes import CalcCarbonIsotopes
from pyrealm.pmodel.pmodel import PModel, SubdailyPModel
from pyrealm.pmodel.pmodel_environment import PModelEnvironment

__all__ = [
    "AcclimationModel",
    "C3C4Competition",
    "CalcCarbonIsotopes",
    "JmaxLimitation",
    "PModel",
    "PModelEnvironment",
    "SubdailyPModel",
    "calc_co2_to_ca",
    "calc_ftemp_inst_rd",
    "calc_ftemp_inst_vcmax",
    "calc_gammastar",
    "calc_kmm",
    "calc_ns_star",
    "calc_soilmstress_mengoli",
    "calc_soilmstress_stocker",
    "calc_viscosity_h2o",
    "calculate_kattge_knorr_arrhenius_factor",
    "calculate_simple_arrhenius_factor",
    "calculate_tree_proportion",
    "convert_gpp_advantage_to_c4_fraction",
]



"""The pmodel_const module TODO."""

from dataclasses import dataclass, field

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import ConstantsClass


@dataclass(frozen=True)
class PModelConst(ConstantsClass):
    r"""Constants for the P Model module.

    This dataclass provides constants used in calculating the predictions of the P Model
    and associated methods.
    """

    sandoval_peak_phio: tuple[float, float] = (6.8681, 0.07956432)
    """Curvature parameters for calculation of peak phio in the Sandoval method for
    estimation of quantum yield efficiency."""

    sandoval_kinetics: dict[str, float] = field(
        default_factory=lambda: dict(
            entropy_intercept=1558.853,
            entropy_slope=-50.223,
            ha=75000.0,
            hd=294.804,
        )
    )
    """Enzyme kinetics parameters for estimation of kphio from mean growth temperature
    in the Sandoval method :cite:t:`sandoval:in_prep` for estimation of quantum yield
    efficiency. Values are: the intercept and slope of activation entropy as a function
    of the mean growth temperature (J/mol/K), the deactivation energy constant (J/mol)
    and the activation energy (J/mol). """

    tc_ref: float = 25.0
    """Standard baseline reference temperature of photosynthetic processes in °C 
    (:math:`T_o` , 25.0, °C)"""

    tk_ref: float = 298.15
    """Standard baseline reference temperature of photosynthetic processes in Kelvin 
    (298.15, K)"""

    # NOTE that tk_ref might be a considered a duplication of CoreConst.k_To but that
    # refers explicitly to the physical standard temperature and this is a reference
    # temperature for photosynthesis. It is _vanishingly_ unlikely that this will
    # change but they could _in theory_ be different.

    heskel_rd: tuple[float, float] = (0.1012, 0.0005)
    """Linear (:math:`b`, 0.1012) and quadratic  (:math:`c`, 0.0005) coefficients of the
    temperature scaling of dark respiration. Values taken from
    :cite:t:`Heskel:2016fg`:."""

    # Arrhenius values
    arrhenius_vcmax: dict = field(
        default_factory=lambda: dict(
            simple=dict(ha=65330),
            kattge_knorr=dict(
                entropy_intercept=668.39,
                entropy_slope=-1.07,
                ha=71513,
                hd=200000,
            ),
        )
    )
    """Coefficients of Arrhenius factor scaling for :math:`V_{cmax}`. The ``simple``
    method provides an estimate of the activation energy (:math:`H_a`, J/mol). The
    ``kattge_knorr`` method provides the parameterisation from :cite:t:`Kattge:2007db`, 
    providing the intercept and slope of activation entropy as a function of the mean
    growth temperature (J/mol/K), the deactivation energy constant (:math:`H_d`, J/mol) 
    and the activation energy (:math:`H_a`, J/mol)."""

    arrhenius_jmax: dict = field(
        default_factory=lambda: dict(
            simple=dict(ha=43900),
            kattge_knorr=dict(
                entropy_intercept=659.70,
                entropy_slope=-0.75,
                ha=49884,
                hd=200000,
            ),
        )
    )
    """Coefficients of Arrhenius factor scaling for :math:`J_{max}`. The ``simple``
    method provides an estimate of the activation energy (:math:`H_a`, J/mol). The
    ``kattge_knorr`` method provides the parameterisation from :cite:t:`Kattge:2007db`, 
    providing the intercept and slope of activation entropy as a function of the mean
    growth temperature (J/mol/K), the deactivation energy constant (:math:`H_d`, J/mol) 
    and the activation energy (:math:`H_a`, J/mol)."""

    kphio_C4: tuple[float, float, float] = (-0.064, 0.03, -0.000464)
    r"""Coefficients of the quadratic scaling of the quantum yield of photosynthesis
    (``phi_0``, :math:`\phi_0`) with temperature for C4 plants, taken from Eqn 5 of
    :cite:t:`cai:2020a`, and adjusted from the original values (-0.008, 0.00375,
    -0.58e-4) to account for an unintentional double scaling to account for the fraction
    of light reaching PS2."""

    kphio_C3: tuple[float, float, float] = (0.352, 0.022, -0.00034)
    r"""Coefficients of the quadratic scaling of the quantum yield of photosynthesis
    (``phi_0``, :math:`\phi_0`) with temperature for C3 plants, taken from Table 2 of
    :cite:t:`Bernacchi:2003dc`"""

    # Bernachhi
    bernacchi_kmm: dict[str, float] = field(
        default_factory=lambda: dict(
            dhac=79430.0,
            dhao=36380.0,
            kc25=39.97,
            ko25=27480.0,
        )
    )
    r"""Coefficients for the estimation of the Michaelis Menten coefficient of
    Rubisco-limited assimilation. Values are taken from Table 1 of
    :cite:t:`Bernacchi:2001kg` and provide a dictionary of:
    
    * The activation energy for :math:`\ce{CO2}` (:math:`\Delta H_{kc}`, ``dhac``,
      J/mol) 
    * The activation energy for :math:`\ce{O2}` (:math:`\Delta H_{ko}`, `dhao``, J/mol))
    * The Michelis constant for :math:`\ce{CO2}` at standard temperature
      (:math:`K_{c25}`, ``kc25``, Pa) 
    * The Michelis constant for :math:`\ce{O2}` at standard temperature
      (:math:`K_{o25}`, ``ko25``, Pa)

    The values for  `kc_25` and `ko_25` are converted from the original tabulated values
    of 404.9 µmol mol-1 and 278.4 mmol mol-1 respectively to Pa, assuming a measurement
    at an elevation of 227.076 metres and standard  atmospheric pressure for that
    elevation (98716.403 Pa).
    """

    maximum_phi0: float = 1.0 / 8.0
    r"""The theoretical maximum intrinsic quantum yield efficiency of photosynthesis
    (:math:`\phi_0`, unitless). This value defaults to 1/8 but could be set to 1/9 to
    capture the absence of a Q cycle :cite:`long:1993a`."""

    bernacchi_gs: dict[str, float] = field(
        default_factory=lambda: dict(
            dha=37830.0,
            gs25_0=4.332,
        )
    )
    r"""Coefficients for the estimation of the photorespiratory CO2 compensation point.
    Values are taken from Table 1 of :cite:t:`Bernacchi:2001kg` and provide a dictionary
    of: 
    
    * The reference value of :math:`\Gamma^{*}` at standard temperature and pressure
      (:math:`\Gamma^{*}_{0}`, ``gs25_0``, Pa), converted from the tabulated value of
      42.75 µmol mol-1, , assuming a measurement at an elevation of 227.076 metres and
      standard  atmospheric pressure for that elevation (98716.403 Pa).
    * The activation energy (:math:`\Delta H_a`, ``dha``, J/mol)

    The values for  `kc_25` and `ko_25` are converted from the original tabulated values
    of 404.9 µmol mol-1 and 278.4 mmol mol-1 respectively to Pa, assuming a measurement
    at an elevation of 227.076 metres and standard  atmospheric pressure for that
    elevation (98716.403 Pa).
    """

    soilmstress_stocker: dict[str, float] = field(
        default_factory=lambda: dict(theta0=0, thetastar=0.6, a=0.0, b=0.733)
    )
    """Parameterisation of the soil moisture stress function of
    :cite:t:`Stocker:2020dh` as a dictionary providing values for the: 
    
    * intercept of the aridity sensitivity function (``theta0``),
    * slope of the aridity sensitivity function (``thetastar``),
    * lower bound in relative soil moisture  (``a``), and
    * upper bound in relative soil moisture (``b``).
    """

    soilmstress_mengoli: dict[str, float] = field(
        default_factory=lambda: dict(psi_a=0.34, psi_b=-0.6, y_a=0.62, y_b=-0.45)
    )
    """Parameterisation of the soil moisture stress function of
    :cite:t:`mengoli:2023a` as a dictionary providing values for the:

    * coefficient of the maximal level function  (``y_a``),
    * exponent of the maximal level function  (``y_b``),
    * coefficient of the threshold function  (``psi_a``), and
    * exponent of the threshold function (``psi_b``).
    """

    beta_cost_ratio_c3: NDArray[np.float64] = field(
        default_factory=lambda: np.array([146.0])
    )
    r"""Unit cost ratio for C3 plants (:math:`\beta`, 146.0) taken from
     :cite:t:`Stocker:2020dh`."""

    beta_cost_ratio_c4: NDArray[np.float64] = field(
        default_factory=lambda: np.array([146.0 / 9])
    )
    r"""Unit cost ratio for C4 plants (:math:`\beta`, 16.222).
    
    For C4 plants, we follow the estimates of the   :math:`g_1` parameter for C3 and C4
    plants in :cite:t:`Lin:2015wh`  and :cite:t:`DeKauwe:2015im`, which have a C3/C4
    ratio of around 3. Given that :math:`g_1 \equiv \xi \propto \surd\beta`, a
    reasonable default for C4 plants is that :math:`\beta_{C4} \approx \beta_{C3} / 9
    \approx 146 /  9 \approx 16.222`."""

    lavergne_2020_c3: tuple[float, float] = (4.55, 1.73)
    """Intercept and slope coefficients for the effects of soil moisture on optimal chi
    estimates for C3 plants, following :cite:`lavergne:2020a`."""

    lavergne_2020_c4: tuple[float, float] = (4.55 - np.log(9), 1.73)
    """Intercept and slope coefficients for the effects of soil moisture on optimal chi
    estimates for C4 plants, following :cite:`lavergne:2020a`."""

    wang17_c: float = 0.41
    """Unit carbon cost for the maintenance of electron transport capacity, value taken
    from  :cite:t:`Wang:2017go` (:math:`c`, 0.41, )."""

    # Smith19
    smith19_coef: tuple[float, float] = (0.85, 0.05336251)
    r"""Scaling factors ``theta`` (:math:`\theta`, 0.85) and ``c`` (:math:`c`,
    0.05336251) used in the calculation of J_max limitation using the method of 
    :cite:t:`Smith:2019dv`."""



"""Arrhenius enzyme reaction models.

The reaction rates of enzyme systems of photosynthesis respond to temperature in a way
that is estimate using an Arrhenius relationship. There are alternative ways of
parameterizing this temperature scaling and this module defines functions implementing
different approaches to calculating Arrhenius factors for a given temperature, relative
to a reference temperature, given a set of coefficients for a specific enzyme system.

It also provides the :class:`pyrealm.pmodel.arrhenius.ArrheniusFactorABC` abstract base
class. This provides a framework that allows different scaling functions to be called
from the PModel and SubdailyPModel classes, and hence allows the models to use different
temperature relationships for enzyme rates.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from numpy.typing import NDArray

from pyrealm.core.experimental import warn_experimental
from pyrealm.pmodel.functions import (
    calculate_kattge_knorr_arrhenius_factor,
    calculate_simple_arrhenius_factor,
)
from pyrealm.pmodel.pmodel_environment import PModelEnvironment

ARRHENIUS_METHOD_REGISTRY: dict[str, type[ArrheniusFactorABC]] = {}
"""A registry for Arrhenius factor calculation classes.

This dictionary is used as a registry for subclasses of the
:class:`~pyrealm.pmodel.arrhenius.ArrheniusFactorABC` abstract base class. The different
subclasses are keyed by a method name that can be used to retrieve a particular
implementation from this registry. For example: 

.. code:: python

    arrh_simple = ARRHENIUS_METHOD_REGISTRY['simple']
"""


class ArrheniusFactorABC(ABC):
    """Abstract base class for implementations of Arrhenius factors.

    This abstract base class provides a framework for implementing Arrhenius
    calculations within a PModel or SubdailyPModel. Individual subclasses are registered
    by a method name that can then be used with the ``method_arrhenius`` argument to
    those classes.

    The ``__init__`` method uses the PModelEnvironment to provide temperature and any
    other required variables to the calculation, along with the reference temperature to
    be used. The ``calculate_arrhenius_factor`` method provides validation to check that
    the coefficients required by a particular implementation are provided.

    Subclasses only need to implement the private abstract method `_calculate_values`,
    which should implement the actual calculation from the data and coefficients
    supplied to the instance, and should return an array of calculated factor values.
    """

    method: str
    """A short method name used to identify the class in
    :data:`~pyrealm.pmodel.arrhenius.ARRHENIUS_METHOD_REGISTRY`. This name is also used
    to select a matching dictionary of coefficients from inputs.
    """
    required_coefficients: set[str]
    """A set of the names of coefficients required to calculate factor values. These
    must be present in the coefficients dictionary matching the method name."""

    required_env_variables: list[str]
    """A list of names of optional attributes of
    :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment` that must be populated
    to use a method.
    """

    def __init__(
        self,
        env: PModelEnvironment,
    ):
        self.env: PModelEnvironment = env
        """The PModelEnvironment containing the photosynthetic environment for the
        model."""

        # Run the calculation methods after checking for any required variables
        self._check_required_env_variables()

    @abstractmethod
    def _calculation_method(self, coefficients: dict) -> NDArray:
        pass

    def calculate_arrhenius_factor(self, coefficients: dict) -> NDArray:
        """Calculate the Arrhenius factor.

        This method calculates the Arrhenius factor for the model environment, given a
        dictionary providing the required enzyme coefficients.

        Args:
            coefficients: A dictionary providing any required coefficients for a given
                enzyme system.

        Raises:
            ValueError: where the method name is not found in the coefficients
                dictionary or the required coefficents are not found in that matched
                dictionary.
        """

        # Check the coefficients dictionary provides an entry for this method
        if self.method not in coefficients:
            raise ValueError(
                f"The coefficients dict does not provide a parameterisation "
                f"for the {self.method} Arrhenius method."
            )

        # Check the required coefficients are found
        missing_coefficients = self.required_coefficients.difference(
            coefficients[self.method]
        )
        if missing_coefficients:
            raise ValueError(
                f"The coefficients for the {self.method} Arrhenius method do not "
                f"provide: {','.join(sorted(missing_coefficients))}"
            )

        return self._calculation_method(coefficients=coefficients[self.method])

    def _check_required_env_variables(self) -> None:
        """Check additional required variables are present."""

        for required_var in self.required_env_variables:
            if not hasattr(self.env, required_var):
                raise ValueError(
                    f"{self.__class__.__name__} (method {self.method}) requires "
                    f"{required_var} to be provided in the PModelEnvironment."
                )

    @classmethod
    def __init_subclass__(
        cls,
        method: str,
        required_coefficients: set[str],
        required_env_variables: list[str],
    ) -> None:
        """Initialise a subclass deriving from this ABC."""

        cls.method = method
        cls.required_coefficients = required_coefficients
        cls.required_env_variables = required_env_variables

        ARRHENIUS_METHOD_REGISTRY[cls.method] = cls


class SimpleArrhenius(
    ArrheniusFactorABC,
    method="simple",
    required_coefficients={"ha"},
    required_env_variables=[],
):
    """Class providing simple Arrhenius scaling.

    This class provides an implementation of simple Arrhenius scaling for the data in a
    PModelEnvironment. It requires no variables other than the standard temperature and
    requires that a coefficient dictionary providing only ``ha`` (the activation energy
    constant, :math:`H_a`, J/mol).

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]),
        ...     patm=np.array([101325]),
        ...     co2=np.array([400]),
        ...     vpd=np.array([1000]),
        ... )
        >>> arrh = SimpleArrhenius(env=env)
        >>> # Simple Arrhenius scaling factor using V_cmax coefficients
        >>> arrh.calculate_arrhenius_factor(
        ...     coefficients={'simple': {'ha': 65330}}
        ... ).round(5)
        array([0.63795])
    """

    def _calculation_method(self, coefficients: dict) -> NDArray:
        return calculate_simple_arrhenius_factor(
            tk=self.env.tk,
            tk_ref=self.env.pmodel_const.tk_ref,
            ha=coefficients["ha"],
            k_R=self.env.core_const.k_R,
        )


class KattgeKnorrArrhenius(
    ArrheniusFactorABC,
    method="kattge_knorr",
    required_coefficients={"ha", "hd", "entropy_intercept", "entropy_slope"},
    required_env_variables=["mean_growth_temperature"],
):
    """Class providing Kattge Knorr Arrhenius scaling.

    This method implements the peaked Arrhenius scaling model of
    :cite:t:`Kattge:2007db`. It requires that the PModelEnvironment also provides values
    for the mean growth temperature of plants as ``mean_growth_temperature`` in °C. It
    also requires a coefficients dictionary providing:

    * the intercept (``entropy_intercept``) and slope (``entropy_slope``) of activation
      entropy as a function of the mean growth temperature in °C (J/mol/°C),
    * the deactivation energy constant  (``hd``, :math:`H_d`, J/mol) and
    * the activation energy constant (``ha``, :math:`H_a`, J/mol).

    Examples:
        >>> import numpy as np
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]),
        ...     patm=np.array([101325]),
        ...     co2=np.array([400]),
        ...     vpd=np.array([1000]),
        ...     mean_growth_temperature=np.array([10]),
        ... )
        >>> arrh = KattgeKnorrArrhenius(env=env)
        >>> # Kattge and Knorr Arrhenius scaling factor using V_cmax coefficients
        >>> arrh.calculate_arrhenius_factor(
        ...     coefficients={"kattge_knorr":
        ...         {
        ...             'entropy_intercept': 668.39,
        ...             'entropy_slope': -1.07,
        ...             'ha': 71513,
        ...             'hd': 200000,
        ...         }
        ...     }
        ...  ).round(5)
        array([0.70109])
    """

    __experimental__ = True

    def _calculation_method(self, coefficients: dict) -> NDArray:
        warn_experimental("KattgeKnorrArrhenius")

        return calculate_kattge_knorr_arrhenius_factor(
            tk_leaf=self.env.tk,
            tk_ref=self.env.pmodel_const.tk_ref,
            tc_growth=getattr(self.env, "mean_growth_temperature"),
            coef=coefficients,
            k_R=self.env.core_const.k_R,
        )



"""The :mod:`~pyrealm.pmodel.jmax_limitation` module provides the implementation of
classes for calculation of :math:`J_{max}` and :math:`V_{cmax}` limitation. The module
provides an abstract base dataclass
(:class:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationABC`) which provides the core
functionality for the implementation. Individual methods then are defined as subclasses
that only need to add any additional data attributes and define the private
:meth:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationABC._calculate_limitation_terms`
method. This is automatically called by the ``__post_init__`` method of the data class and
so the limitation terms are calculated when an instance is created.

The module defines a registry
(:data:`~pyrealm.pmodel.jmax_limitation.JMAX_LIMITATION_CLASS_REGISTRY`) to track
defined subclasses. Subclasses are added to this dictionary, under a string set by the
subclass ``method`` attribute, by the ``__init_subclass`` method of the base class,
which allows implementations to be selected by a simple string method name.
"""  # noqa D210, D415

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import ClassVar

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import PModelConst
from pyrealm.core.utilities import summarize_attrs
from pyrealm.pmodel.optimal_chi import OptimalChiABC

JMAX_LIMITATION_CLASS_REGISTRY: dict[str, type[JmaxLimitationABC]] = {}
"""A registry for subclasses of
:class:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationABC`. Subclasses are automatically
included in this registry dictionary under their defined ``method`` name.
"""


@dataclass
class JmaxLimitationABC(metaclass=ABCMeta):
    r"""An abstract base class for JMaxLimitation implementations.

    This base class defines the ``__init__`` arguments, common data attributes and core
    methods for implementing JMaxLimitation methods. Subclasses should only need to
    define any additional data attributes that should be exposed to users and define the
    private
    :meth:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationABC._calculate_limitation_terms`
    method for the implementation. Subclass definitions should use
    ``@dataclass(repr=False)`` to avoid overriding the base implementation of the
    ``_repr__`` method, and also need to provide a method name string and a tuple of the
    data attributes to include when the
    :meth:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationABC.summarize` method is called
    for the subclass.

    See :class:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationWang17` for an example.
    """

    method: ClassVar[str]
    """A short name for the method of Jmax limitation implemented in the subclass."""
    data_attrs: ClassVar[tuple[tuple[str, str], ...]]
    """A tuple of names and units for the data attributes of the class to be reported 
    by summarize."""
    optchi: OptimalChiABC
    """The optimal chi instance used to calculate limitation terms."""
    pmodel_const: PModelConst = field(default_factory=lambda: PModelConst())
    """The PModel constants instance used for the calculation."""
    _shape: tuple[int, ...] = field(init=False)
    """Records the common numpy array shape in the data."""
    f_j: NDArray[np.float64] = field(init=False)
    """:math:`J_{max}` limitation factor."""
    f_v: NDArray[np.float64] = field(init=False)
    """:math:`V_{cmax}` limitation factor."""

    def __post_init__(self) -> None:
        self._shape = self.optchi.mj.shape

        self._calculate_limitation_terms()

    def __repr__(self) -> str:
        """Generates a string representation of a JmaxLimitation instance."""
        return f"JmaxLimitation(method={self.method}, shape={self._shape})"

    def summarize(self, dp: int = 2) -> None:
        """Print OptimalChi summary.

        Prints a summary of the variables calculated within an instance
        of OptimalChi including the mean, range and number of nan values.

        Args:
            dp: The number of decimal places used in rounding summary stats.
        """

        summarize_attrs(self, self.data_attrs, dp=dp)

    @abstractmethod
    def _calculate_limitation_terms(self) -> None:
        """Abstract method defined in subclasses to populate limitation attributes."""
        pass

    @classmethod
    def __init_subclass__(
        cls, method: str, data_attrs: tuple[tuple[str, str], ...]
    ) -> None:
        """Initialise a subclass deriving from this ABC."""

        cls.method = method
        cls.data_attrs = data_attrs
        JMAX_LIMITATION_CLASS_REGISTRY[cls.method] = cls


@dataclass(repr=False)
class JmaxLimitationWang17(
    JmaxLimitationABC,
    method="wang17",
    data_attrs=(("f_j", "-"), ("f_v", "-")),
):
    r"""Calculate limitation factors following :cite:`Wang:2017go`.

    These factors are described in Equation 49 of :cite:`Wang:2017go` as the
    square root term at the end of that equation:

        .. math::
            :nowrap:

            \[
                \begin{align*}
                f_v &=  \sqrt{ 1 - \frac{c^*}{m} ^{2/3}} \\
                f_j &=  \sqrt{\frac{m}{c^*} ^{2/3} -1 } \\
                \end{align*}
            \]

    The variable :math:`c^*` is a cost parameter for maintaining :math:`J_{max}`
    and is set in
    :attr:`PModelConsts.wang17_c<pyrealm.constants.PModelConst.wang17_c>`. 
    Note that both equations are undefined where :math:`m \le c^*`: where this
    condition is true, values will be returned as ``np.nan``.

    Examples:
        >>> from pyrealm.pmodel import PModelEnvironment
        >>> from pyrealm.pmodel.optimal_chi import OptimalChiPrentice14
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ... )
        >>> optchi = OptimalChiPrentice14(env=env)
        >>> jmaxlim = JmaxLimitationWang17(optchi=optchi)
        >>> jmaxlim.f_j.round(4)
        array([0.6672])
        >>> jmaxlim.f_v.round(4)
        array([0.555])
    """

    def _calculate_limitation_terms(self) -> None:
        """Limitation calculations for the ``wang17`` method."""
        # Test for m > c*
        vals_defined = np.greater(self.optchi.mj, self.pmodel_const.wang17_c)

        self.f_v = np.sqrt(
            1 - (self.pmodel_const.wang17_c / self.optchi.mj) ** (2.0 / 3.0),
            where=vals_defined,
        )
        self.f_j = np.sqrt(
            (self.optchi.mj / self.pmodel_const.wang17_c) ** (2.0 / 3.0) - 1,
            where=vals_defined,
        )

        # Backfill undefined values
        self.f_j[np.logical_not(vals_defined)] = np.nan
        self.f_v[np.logical_not(vals_defined)] = np.nan


@dataclass(repr=False)
class JmaxLimitationSmith19(
    JmaxLimitationABC,
    method="smith19",
    data_attrs=(("f_j", "-"), ("f_v", "-"), ("omega", "-"), ("omega_star", "-")),
):
    r"""Calculate limitation factors following :cite:`Smith:2019dv`.

    The values are calculated as:

    .. math::
        :nowrap:

        \[
            \begin{align*}
            f_v &=  \frac{\omega^*}{2\theta} \\
            f_j &=  \omega\\
            \end{align*}
        \]

    where,

    .. math::
        :nowrap:

        \[
            \begin{align*}
            \omega &= (1 - 2\theta) + \sqrt{(1-\theta)
                \left(\frac{1}{\frac{4c}{m}(1 -
                \theta\frac{4c}{m})}-4\theta\right)}\\
            \omega^* &= 1 + \omega - \sqrt{(1 + \omega) ^2 -4\theta\omega}
            \end{align*}
        \]

    given,

    * :math:`\theta`, (``const.smith19_theta``) captures the
        curved relationship between light intensity and photosynthetic
        capacity, and
    * :math:`c`, (``const.smith19_c_cost``) as a cost parameter
        for maintaining :math:`J_{max}`, equivalent to :math:`c^\ast = 4c`
        in the :class:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationWang17` limitation
        terms.

    Examples:
        >>> from pyrealm.pmodel import PModelEnvironment
        >>> from pyrealm.pmodel.optimal_chi import OptimalChiPrentice14
        >>> env = PModelEnvironment(
        ...     tc=np.array([20]), vpd=np.array([1000]),
        ...     co2=np.array([400]), patm=np.array([101325.0]),
        ...     fapar=np.array([1]), ppfd=np.array([800]),
        ... )
        >>> optchi = OptimalChiPrentice14(env=env)
        >>> jmaxlim = JmaxLimitationSmith19(optchi=optchi)
        >>> jmaxlim.f_j.round(4)
        array([1.102])
        >>> jmaxlim.f_v.round(4)
        array([0.7544])
    """

    omega: NDArray[np.float64] = field(init=False)
    """Values of the `omega` parameter (:cite:`Smith:2019dv`)."""
    omega_star: NDArray[np.float64] = field(init=False)
    """Values of the `omega_star` parameter (:cite:`Smith:2019dv`)."""

    def _calculate_limitation_terms(self) -> None:
        """Limitation calculations for the ``smith19`` method."""

        # Calculate omega, see Smith et al., 2019 Ecology Letters  # Eq. S4
        theta, c_cost = self.pmodel_const.smith19_coef

        # simplification terms for omega calculation
        cm = 4 * c_cost / self.optchi.mj
        v = 1 / (cm * (1 - theta * cm)) - 4 * theta

        # account for non-linearities at low m values. This code finds
        # the roots of a quadratic function that is defined purely from
        # the scalar theta, so will always be a scalar. The first root
        # is then used to set a filter for calculating omega.

        cap_p = (((1 / 1.4) - 0.7) ** 2 / (1 - theta)) + 3.4
        aquad = -1
        bquad = cap_p
        cquad = -(cap_p * theta)
        roots = np.polynomial.polynomial.polyroots([aquad, bquad, cquad])  # type: ignore [no-untyped-call]

        # factors derived as in Smith et al., 2019
        m_star = (4 * c_cost) / roots[0].real
        omega = np.where(
            self.optchi.mj < m_star,
            -(1 - (2 * theta)) - np.sqrt((1 - theta) * v),
            -(1 - (2 * theta)) + np.sqrt((1 - theta) * v),
        )

        # np.where _always_ returns an array, so catch scalars
        self.omega = omega.item() if np.ndim(omega) == 0 else omega

        self.omega_star = (
            1.0
            + self.omega
            - np.sqrt((1.0 + self.omega) ** 2 - (4.0 * theta * self.omega))  # Eq. 18
        )

        # Effect of Jmax limitation - note scaling here. Smith et al use
        # phi0 as as the quantum efficiency of electron transport, which is
        # 4 times our definition of phio0 as the quantum efficiency of photosynthesis.
        # So omega*/8 theta and omega / 4 are scaled down here  by a factor of 4.
        self.f_v = self.omega_star / (2.0 * theta)
        self.f_j = self.omega


@dataclass(repr=True)
class JmaxLimitationNone(
    JmaxLimitationABC, method="none", data_attrs=(("f_j", "-"), ("f_v", "-"))
):
    """No limitation of :math:`J_{max}` and :math:`V_{cmax}`.

    This implementation simply sets :math:`f_v = f_j = 1` to remove any :math:`J_{max}`
    and :math:`V_{cmax}` limitation.
    """

    def _calculate_limitation_terms(self) -> None:
        """Set limitation terms to one."""

        # Set limitation terms to unity
        self.f_v = np.ones(self._shape)
        self.f_j = np.ones(self._shape)
